import paicos as pa

"""
Set up your own default settings by renaming this file as user_settings.py.

Here we are overriding the defaults set in settings.py, so you only need
to add things you want to change.
"""

# Boolean determining whether we use Paicos quantities as a default
pa.use_units(True)

# Settings for automatic calculation of derived variables
pa.use_only_user_functions(False)
pa.print_info_when_deriving_variables(False)

# Number of threads to use in calculations
pa.numthreads(8)

# Info about the openMP setup
pa.give_openMP_warnings(False)

# Whether to load GPU/cuda functionality on startup
pa.load_cuda_functionality_on_startup(False)

# Explicitly set data directory (only needed for pip installations)
data_dir = '/Users/berlok/projects/paicos/data/'


# Examples of adding user-defined functions
def TemperaturesTimesMassesSquared(snap, get_dependencies=False):
    if get_dependencies:
        return ['0_TemperaturesTimesMasses', '0_Masses']
    return snap['0_TemperaturesTimesMasses'] * snap['0_Masses']


pa.add_user_function('0_TM2', TemperaturesTimesMassesSquared)



"""
Defines a class that creates an image of a given variable by projecting
the Voronoi cells closest to the line of sight using a KDTree.
"""
import numpy as np
from scipy.spatial import KDTree
from .image_creator import ImageCreator
from .. import util
from .. import settings


class TreeProjector(ImageCreator):
    """
    This implements projection of gas variables by adding up several slices.
    """

    def __init__(self, snap, center, widths, direction,
                 npix=512, npix_depth=None, parttype=0, make_snap_with_selection=False,
                 tol=1, verbose=False):

        """
        Initialize the Slicer object.

        Parameters
        ----------
        snap : Snapshot
            A snapshot object of Snapshot class from paicos package.

        center :
            Center of the region on which projection is to be done, e.g.
            center = [x_c, y_c, z_c].

        widths :
            Widths of the region on which projection is to be done,
            e.g. widths=[width_x, width_y, width_z].

        direction : str
            Direction of the projection, e.g. 'x', 'y' or 'z'.

        npix : int, optional
            Number of pixels in the horizontal direction of the image,
            by default 512.

        parttype : int, optional
            The particle type to project, by default gas (PartType 0).

        npix_depth: int, optional
            Number of pixels in the depth direction, by default set
            automatically based on the smallest cell sizes in the region
            and the tolerance parameter, tol (see below).

        tol: float, optional
            Smaller values of tol adds more slices to the integration.
            Convergence is expected for tol ≤ 1 but you can experiment with
            higher values.

        make_snap_with_selection : bool
            a boolean indicating if a new snapshot object should be made with
            the selected region, defaults to False

        """

        if make_snap_with_selection:
            raise RuntimeError('make_snap_with_selection not yet implemented!')

        super().__init__(snap, center, widths, direction, npix=npix, parttype=parttype)

        self.npix_depth = npix_depth
        self.tol = tol
        self.verbose = verbose

        self._do_region_selection()

    def _do_region_selection(self):

        self.do_unit_consistency_check()

        # print('_do_region_selection was called from Slicer')
        parttype = self.parttype
        snap = self.snap
        center = self.center
        widths = self.widths

        # Pre-select a narrow region around the region-of-interest
        avail_list = (list(snap.keys()) + snap._auto_list)
        if f'{parttype}_Volume' in avail_list:
            thickness = 4.0 * np.cbrt((snap[f"{parttype}_Volume"]) / (4.0 * np.pi / 3.0))
        elif f'{parttype}_SubfindHsml' in avail_list:
            thickness = snap[f'{parttype}_SubfindHsml']
        else:
            err_msg = ("There is no smoothing length or volume for calculating"
                       + "the thickness of the slice")
            raise RuntimeError(err_msg)

        if hasattr(thickness, 'unit'):
            thickness = thickness.to(center.unit)

        if self.direction != 'orientation':
            get_index = util.get_index_of_cubic_region_plus_thin_layer
            self.box_selection = get_index(snap[f"{parttype}_Coordinates"],
                                           center, widths, thickness,
                                           snap.box)
        else:
            get_index = util.get_index_of_rotated_cubic_region_plus_thin_layer
            self.box_selection = get_index(snap[f"{parttype}_Coordinates"],
                                           center, widths, thickness, snap.box,
                                           self.orientation)

        if self.verbose:
            print('Sub-selection [DONE]')

        min_thickness = np.min(thickness)

        # Automatically set numbers of pixels in depth direction based on cell sizes
        if self.npix_depth is None:
            npix_depth = int(np.ceil(self.depth / min_thickness / self.tol))
            if self.verbose:
                print(f'npix_depth is {npix_depth}')

            self.npix_depth = npix_depth

        self.index_in_box_region = np.arange(snap[f"{self.parttype}_Coordinates"].shape[0]
                                             )[self.box_selection]

        # Construct a tree
        self.pos = snap[f"{parttype}_Coordinates"][self.box_selection]
        tree = KDTree(self.pos)
        if self.verbose:
            print('Tree construction [DONE]')

        # Now construct the image grid
        w, h = self._get_width_and_height_arrays()

        center = self.center
        ones = np.ones(w.shape[0])

        # Full index
        self.index = np.empty((self.npix_height, self.npix_width, self.npix_depth),
                              dtype=np.int64)

        self.delta_depth = self.depth / self.npix_depth
        depth_vector = np.arange(self.npix_depth) * self.delta_depth \
            + self.delta_depth / 2 - self.depth / 2

        for ii, dep in enumerate(depth_vector):
            if self.direction == 'x':
                image_points = np.vstack([ones * (center[0] + dep), w, h]).T
            elif self.direction == 'y':
                image_points = np.vstack([h, ones * (center[1] + dep), w]).T
            elif self.direction == 'z':
                image_points = np.vstack([w, h, ones * (center[2] + dep)]).T
            elif self.direction == 'orientation':
                orientation = self.orientation
                image_points = np.vstack([w, h, ones * (center[2] + dep)]).T - self.center
                image_points = np.matmul(orientation.rotation_matrix, image_points.T).T \
                    + self.center
            else:
                raise RuntimeError(f"Problem with direction={self.direction} input")

            # Query the tree to obtain closest Voronoi cell indices
            d, i = tree.query(image_points, workers=settings.numthreads)

            slice_index = self._unflatten(self.index_in_box_region[i])
            self.distance_to_nearest_cell = self._unflatten(d)

            if min_thickness < self.delta_depth / self.tol:
                print(f'Warning: Minimum cell size {min_thickness} is '
                      + f'less than {self.tol} of delta_depth '
                      + f'{self.delta_depth}. You should probably increase '
                      + f'npix_depth from its current value of {self.npix_depth}. '
                      + 'Image convergence is expected for npix_depth='
                      + f'{int(self.depth/min_thickness)}.')

            self.index[:, :, ii] = slice_index

    def _get_width_and_height_arrays(self):
        """
        Get width and height coordinates in the image as 1D arrays
        of total length npix_width × npix_height.
        """

        # TODO: Make this part of the image_creator class
        extent = self.extent

        npix_width = self.npix_width
        npix_height = self.npix_height
        width = self.width
        height = self.height

        w = extent[0] + (np.arange(npix_width) + 0.5) * width / npix_width
        h = extent[2] + (np.arange(npix_height) + 0.5) * height / npix_height

        if settings.use_units:
            wu = w.unit_quantity
            ww, hh = np.meshgrid(w.value, h.value)
            ww = ww * wu
            hh = hh * wu
        else:
            ww, hh = np.meshgrid(w, h)

        w = ww.flatten()
        h = hh.flatten()

        np.testing.assert_array_equal(ww, self._unflatten(ww.flatten()))

        return w, h

    def _unflatten(self, arr):
        """
        Helper function to un-flatten 1D arrays to a 2D image
        """
        return arr.flatten().reshape((self.npix_height, self.npix_width))

    def project_variable(self, variable, additive=True, extrinsic=None):
        """
        Project gas variable based on the Voronoi cells closest to each
        line of sight.

        The behavior of this method depends on whether the variable to be
        projected is additive (extrinsic) or not (intrinsic).
        (see e.g. https://en.wikipedia.org/wiki/Intensive_and_extensive_properties).

        For non-additive (intrinsic) properties, e.g. the density ρ, the returned
        projection, P, is

        P = 1/L ∫ ρ dl ,

        where L is the depth of the projection. That is, it is simply the mean
        of a number of slices.

        For additive (extrinsic) properties, e.g. mass M, the returned projection
        is instead

        P = 1 / dA ∫ dM

        where dA is the area per pixel and dM is the mass inside a voxel.

        Parameters
        ----------
        variable: a string or an array of shape (N, )
                  representing the gas variable to project

        additive (bool): A boolean indicating whether the variable to be
                    projected is additive (e.g. Masses, Volumes)
                    or not (e.g. Temperature, density, achieved by
                    setting additive=False). This parameter was previously
                    named 'extrinsic'.

        Returns:
            An array of shape (npix, npix) representing the projected gas variable

            For non-additive (intrinsic) variables, the unit of the projection is
            identical to the unit of the input variable. For additive (extrinsic)
            variables, the projection unit is the input variable unit divided by area.

        Examples
        ----------

        We can compute the projected density in two ways.

        tree_projector = pa.TreeProjector(snap, center, widths, 'z')

        Method I (divide projected mass by projected volume):
            tree_M = tree_projector.project_variable('0_Masses')
            tree_V = tree_projector.project_variable('0_Volume')

            tree_rho = tree_M / tree_V

        Method II (directly project the density):

            rho = tree_projector.project_variable('0_Density', extrinsic=False)

        """
        # This calls _do_region_selection if resolution, Orientation,
        # widths or center changed
        self._check_if_properties_changed()

        if extrinsic is not None:
            import warnings
            warnings.warn("The keyword 'extrinsic' has been replaced by 'additive'."
                          + " The support for 'extrinsic' will be removed eventually.")
            additive = extrinsic

        parttype = self.parttype

        if isinstance(variable, str):
            assert int(variable[0]) == parttype, 'projector uses a different parttype'
            variable = self.snap[variable]
        else:
            if not isinstance(variable, np.ndarray):
                raise RuntimeError('Unexpected type for variable')

        assert len(variable.shape) == 1, 'only scalars can be projected'

        if additive:
            avail_list = (list(self.snap.keys()) + self.snap._auto_list)
            if f'{parttype}_Volume' in avail_list:
                variable_density = variable[self.index] / self.snap[f'{parttype}_Volume'][self.index]
                projection = np.sum(variable_density * self.delta_depth, axis=2) / self.depth
            else:
                err_msg = (f"The volume field for parttype {parttype} is required when"
                           + "using additive=True")
                raise RuntimeError(err_msg)
        else:
            variable = variable[self.index]
            projection = np.sum(variable * self.delta_depth, axis=2) / self.depth

        return projection



"""
Defines hdf5 file writers that can be read with a PaicosReader instance.
"""
import os
import h5py
from .. import util


class PaicosWriter:
    """
    This class writes data to self-documenting hdf5 files.

    It is the base class for the ArepoImage writer.
    """

    def __init__(self, reader_object, basedir,
                 basename="paicos_file", add_snapnum=True, mode='w'):
        """
        The constructor for the `PaicosWriter` class.

        Parameters
        ----------

        reader_object (obj): A PaicosReader object (or Snaphot or Catalog object).

        basedir (file path): The folder where the HDF5 file will be saved.

        basename (str, optional): Base name of the HDF5 file. Defaults
                                  to "paicos_file".

        add_snapnum (bool): Whether to add the snapnum to the HDF5 filename.
                            When True, the filename will be "basename_{:03d}.hdf5",
                            when False it will simply be "basename.hdf5"
                            Defaults to True.

        mode (string): The mode to open the file in, either 'w' for write mode
        or 'r' for read mode. (default: 'w')

        """

        self.reader_object = reader_object
        self.org_filename = reader_object.filename

        self.mode = mode

        snapnum = reader_object.snapnum

        if basedir[-1] != '/':
            basedir += '/'

        if not os.path.exists(basedir):
            os.makedirs(basedir)

        self.basedir = basedir
        self.basename = basename

        name = basename

        if add_snapnum:
            name += f'_{snapnum:03d}.hdf5'
        else:
            name += '.hdf5'
        self.filename = basedir + name
        self.tmp_filename = basedir + 'tmp_' + name

        if mode == 'w':
            util.copy_over_snapshot_information(self.org_filename,
                                                self.tmp_filename, 'w')
            self.write_info_about_org_file()
        else:
            self.perform_consistency_checks()

    def write_info_about_org_file(self):
        """
        This function saves some attributes in org_info.
        """
        with h5py.File(self.tmp_filename, 'r+') as file:
            file.create_group('org_info')
            if hasattr(self.reader_object, 'snapnum'):
                file['org_info'].attrs['snapnum'] = self.reader_object.snapnum
            if hasattr(self.reader_object, 'basesubdir'):
                file['org_info'].attrs['basesubdir'] = self.reader_object.basesubdir
            file['org_info'].attrs['basedir'] = self.reader_object.basedir
            file['org_info'].attrs['basename'] = self.reader_object.basename
            file['org_info'].attrs['filename'] = self.reader_object.filename

    def write_data(self, name, data, data_attrs={}, group=None, group_attrs={}):
        """
        Write a data set to the hdf5 file.

        Parameters
        ----------

        name (str): Name of the data to be written.

        data (obj): The data to be written (e.g. PaicosQuantity, PaicosTimeSeries,
                    Astropy Quantity, numpy array).

        data_attrs (dict): Dictionary of attributes for the data. E.g.
                           {'info': 'Here I have written some extra information
                                    about the data set'}. Defaults to an empty
                                    dictionary.

        group (str, optional): Group in the HDF5 file where the data should
                               be saved. Defaults to None in which case the
                               data set is saved at the top level of the hdf5 file).

        group_attrs (dict, optional): Dictionary of attributes for the group.
                                      Defaults to an empty dictionary.
        """

        # pylint: disable= dangerous-default-value

        if self.mode == 'w':
            filename = self.tmp_filename
        else:
            filename = self.filename

        file = h5py.File(filename, 'r+')

        if self.mode == 'a':
            msg = ('PaicosWriter is in amend mode but {} is already '
                   + 'in the group {} in the hdf5 file {}')
            msg = msg.format(name, group, file.filename)

            if group is None:
                if name in file:
                    raise RuntimeError(msg)
            else:
                if group in file:
                    if name in file[group]:
                        raise RuntimeError(msg)

        # Save the data
        util.save_dataset(file, name, data=data, data_attrs=data_attrs,
                          group=group, group_attrs=group_attrs)

    def perform_extra_consistency_checks(self):
        """
        Perform extra consistency checks. This can be overloaded by
        subclasses.
        """
        # pylint: disable=unnecessary-pass
        pass

    def perform_consistency_checks(self):
        """
        Perform consistency checks when trying to amend a file (to avoid
        saving data at different times, for instance)
        """
        with h5py.File(self.filename, 'r') as f:
            org_time = self.reader_object.Header['Time']
            assert f['Header'].attrs['Time'] == org_time

        self.perform_extra_consistency_checks()

    def finalize(self):
        """
        Move from a temporary to the final filename.
        """
        if self.mode == 'w':
            os.rename(self.tmp_filename, self.filename)


class PaicosTimeSeriesWriter(PaicosWriter):
    """
    Similar to the standard PaicosWriter but here we ensure that
    the snapnum is not part of the resulting filename for the hdf5 file.
    """

    def __init__(self, reader_object, basedir,
                 basename="paicos_time_series", add_snapnum=False, mode='w'):

        super().__init__(reader_object, basedir,
                         basename=basename,
                         add_snapnum=add_snapnum,
                         mode=mode)



import paicos as pa

"""

"""

# Boolean determining whether we use Paicos quantities as a default
pa.use_units(True)

# Settings for automatic calculation of derived variables
pa.use_only_user_functions(False)
pa.print_info_when_deriving_variables(True)

# Number of threads to use in calculations
pa.numthreads(4)

# Info about the openMP setup
pa.give_openMP_warnings(True)

# Whether to load GPU/cuda functionality on startup
pa.load_cuda_functionality_on_startup(False)

# Explicitly set data directory (only needed for pip installations)
data_dir = '/Users/berlok/projects/paicos/data'



"""
This defines a base class for image creators (such as Projector and Slicer)
"""

import numpy as np
import warnings
from .. import util
from .. import settings
from .. import units
from ..orientation import Orientation


class ImageCreator:
    """
    This is a base class for creating images from a snapshot.
    """

    def __init__(self, snap, center, widths, direction, npix=512, parttype=0):
        """
        Initialize the ImageCreator class. This method will be called
        by subclasses such as Projector or Slicer.

        Parameters:
            snap (object): Snapshot object from which the image is created

            center: Center of the image (3D coordinates)

            widths: Width of the image in each direction (3D coordinates)

            direction (str): Direction of the image ('x', 'y', 'z')

            npix (int): Number of pixels in the image (default is 512)
        """

        self._obervers = []

        self.snap = snap

        code_length = self.snap.length

        if hasattr(center, 'unit'):
            self._center = center.copy
            assert center.unit == code_length.unit, 'this restriction applies'
        elif settings.use_units:
            self._center = np.array(center) * code_length
        else:
            self._center = np.array(center)

        # Difference from original center of image
        if settings.use_units:
            self._diff_center = self._center.copy * 0
        else:
            self._diff_center = np.array(self._center) * 0

        if hasattr(widths, 'unit'):
            self._widths = widths.copy
            assert widths.unit == code_length.unit, 'this restriction applies'
        elif settings.use_units:
            self._widths = np.array(widths) * code_length
        else:
            self._widths = np.array(widths)

        if isinstance(direction, str):

            self.direction = direction
            if direction == 'x':
                self.orientation = Orientation(normal_vector=[1, 0, 0],
                                               perp_vector1=[0, 1, 0])
            elif direction == 'y':
                warnings.warn("\n\nThe definition of a 'y'-image was changed "
                              + "on 19/1/24 to be consistent with a right-handed "
                              + "coordinate system. You can recover the old "
                              + "behaviour by passing an orientation object:\n"
                              + "Orientation(normal_vector=[0, -1, 0], perp_vector1=[1, 0, 0])\n"
                              + "instead of a direction 'y' string.\n\n")
                self.orientation = Orientation(normal_vector=[0, 1, 0],
                                               perp_vector1=[0, 0, 1])
            elif direction == 'z':
                self.orientation = Orientation(normal_vector=[0, 0, 1],
                                               perp_vector1=[1, 0, 0])

        elif isinstance(direction, Orientation):
            self.direction = 'orientation'
            self.orientation = direction

        self._npix = self._npix_width = npix

        self._parttype = parttype

        # For checking if properties changes
        self._old_orientation = self.orientation.copy
        self._properties_changed = False

        # Check that units are consistent
        self.do_unit_consistency_check()

        # Check if image region exceeds simulation domain
        # TODO

    def _check_if_properties_changed(self):

        # Check if orientation changed compared to last
        # time _do_region_selection was called
        orientation_changed = not self.orientation._are_equal(self._old_orientation)

        if orientation_changed:
            self.direction == 'orientation'

        # print('_check_if_properties_changed called')
        if self._properties_changed or orientation_changed:
            self._do_region_selection()
            self._properties_changed = False
            self._old_orientation = self.orientation.copy
            return True
        else:
            return False

    def do_unit_consistency_check(self):
        """
        This function is called inside image creators to ensure
        that length units are consistent.
        """

        if settings.use_units:
            if hasattr(self, 'hsml'):
                assert self.hsml.unit.is_equivalent(self.widths.unit)
            if hasattr(self, 'pos'):
                assert self.pos.unit == self.widths.unit, (self.pos.unit, self.widths.unit)
            assert self.center.unit == self.widths.unit, (self.center.unit, self.widths.unit)
            if f"{self.parttype}_Coordinates" in self.snap:
                assert self.center.unit == self.snap[f"{self.parttype}_Coordinates"].unit
            if hasattr(self.snap, 'box_size'):
                assert self.center.unit == self.snap.box_size.unit

    def _do_region_selection(self):
        err_msg = ("_do_region_selection was called from ImageCreator. "
                   + "This should never happen as the Subclasses of "
                   + "ImageCreator should implement this method")
        raise RuntimeError(err_msg)

    def info(self):
        """
        Prints some information about the image creator instance
        TODO: Add things like the number of cells/particles
        in the current instance.
        """
        s = f'Paicos ImageCreator instance of type: {str(self)}\n'
        s += f'widths: {self.widths}\n'
        s += f'center:  {self.center}\n'
        s += f'orientation:  {self.orientation}'
        print(s)

    @property
    def center(self):
        return self._center

    @property
    def widths(self):
        return self._widths

    @widths.setter
    def widths(self, value):
        """
        TODO: ensure that nx, ny
        remain consistent with widths
        """
        assert value.shape[0] == 3
        if settings.use_units:
            assert hasattr(value, 'unit')
            self._widths = value.copy
        else:
            self._widths = np.array(value)
        self._properties_changed = True

    @center.setter
    def center(self, value):
        assert value.shape[0] == 3
        if settings.use_units:
            assert hasattr(value, 'unit')
            self._center = value.copy
        else:
            self._center = np.array(value)
        self._properties_changed = True

    @property
    def npix(self):
        return self._npix

    @npix.setter
    def npix(self, value):
        self._npix = value
        # TODO: Not always needed to do this...
        self._properties_changed = True

    @property
    def npix_width(self):
        return self._npix

    @npix_width.setter
    def npix_width(self, value):
        self.npix = value

    @property
    def parttype(self):
        return self._parttype

    @property
    def foo(self):
        return self._foo

    @property
    def x_c(self):
        return self.center[0]

    @property
    def y_c(self):
        return self.center[1]

    @property
    def z_c(self):
        return self.center[2]

    @property
    def width_x(self):
        return self.widths[0]

    @property
    def width_y(self):
        return self.widths[1]

    @property
    def width_z(self):
        return self.widths[2]

    @property
    def extent(self):
        if self.direction == 'orientation':

            # This is the extent of the image-plane pre-rotation
            # into the orientation of the image...
            # Use centered_extent for showing rotated images.
            center_width = self.x_c
            center_height = self.y_c

        elif self.direction == 'x':
            center_width = self.y_c
            center_height = self.z_c

        elif self.direction == 'y':
            center_width = self.z_c
            center_height = self.x_c

        elif self.direction == 'z':
            center_width = self.x_c
            center_height = self.y_c

        extent = [center_width - self.width / 2, center_width + self.width / 2,
                  center_height - self.height / 2, center_height + self.height / 2]

        if settings.use_units:
            extent = units.paicos_quantity_list_to_array(extent)
        else:
            extent = np.array(extent)
        return extent

    @property
    def centered_extent(self):
        centered_extent = [- self.width / 2, self.width / 2,
                           - self.height / 2, self.height / 2]

        if settings.use_units:
            centered_extent = units.paicos_quantity_list_to_array(centered_extent)
        else:
            centered_extent = np.array(centered_extent)
        return centered_extent

    @property
    def width(self):
        if self.direction == 'x':
            return self.width_y

        elif self.direction == 'y':
            return self.width_z

        elif self.direction == 'z' or self.direction == 'orientation':
            return self.width_x

    @property
    def height(self):
        if self.direction == 'x':
            return self.width_z

        elif self.direction == 'y':
            return self.width_x

        elif self.direction == 'z' or self.direction == 'orientation':
            return self.width_y

    @property
    def depth(self):
        if self.direction == 'x':
            return self.width_x

        elif self.direction == 'y':
            return self.width_y

        elif self.direction == 'z' or self.direction == 'orientation':
            return self.width_z

    @width.setter
    def width(self, value):
        if settings.use_units:
            assert hasattr(value, 'unit')
        if self.direction == 'x':
            self._widths[1] = value.copy

        elif self.direction == 'y':
            self._widths[2] = value.copy

        elif self.direction == 'z' or self.direction == 'orientation':
            self._widths[0] = value.copy
        self._properties_changed = True

    @height.setter
    def height(self, value):
        if settings.use_units:
            assert hasattr(value, 'unit')
        if self.direction == 'x':
            self._widths[2] = value.copy

        elif self.direction == 'y':
            self._widths[0] = value.copy

        elif self.direction == 'z' or self.direction == 'orientation':
            self._widths[1] = value.copy
        self._properties_changed = True

    @depth.setter
    def depth(self, value):
        if settings.use_units:
            assert hasattr(value, 'unit')
        if self.direction == 'x':
            self._widths[0] = value.copy

        elif self.direction == 'y':
            self._widths[1] = value.copy

        elif self.direction == 'z' or self.direction == 'orientation':
            self._widths[2] = value.copy
        self._properties_changed = True

    def double_resolution(self):
        self.npix = self.npix * 2

    def half_resolution(self):
        self.npix = self.npix // 2

    def zoom(self, factor):
        self.width = self.width / factor
        self.height = self.height / factor

        self._properties_changed = True

    def _move_center_along_unit_vector(self, shift, unit_vector):
        """
        Helper function for moving center along unit vector.
        """
        if settings.use_units:
            assert hasattr(shift, 'unit')

        # Calculate new center
        new_center = self._center + shift * unit_vector
        # Keep track of changes to center (offset from initialized center)
        self._diff_center += new_center - self._center

        # Store new center and set bool for re-calculation of image
        self._center = new_center
        self._properties_changed = True

    def move_center_along_normal_vector(self, shift):
        """
        Moves the center of the image along its depth direction.
        Value can be positive or negative and must have same units
        as the center (TODO: In principle we would just convert the distance).
        """
        self._move_center_along_unit_vector(shift, self.orientation.normal_vector)

    def move_center_along_perp_vector1(self, shift):
        """
        Moves the center of the image along its horizontal direction.
        Value can be positive or negative and must have same units
        as the center (TODO: In principle we would just convert the distance).
        """
        self._move_center_along_unit_vector(shift, self.orientation.perp_vector1)

    def move_center_along_perp_vector2(self, shift):
        """
        Moves the center of the image along its vertical direction.
        Value can be positive or negative and must have same units
        as the center (TODO: In principle we would just convert the distance).
        """
        self._move_center_along_unit_vector(shift, self.orientation.perp_vector2)

    @property
    def npix_height(self):
        if settings.use_units:
            return int((self.height / self.width).value * self.npix_width)
        else:
            return int((self.height / self.width) * self.npix_width)

    @property
    def area(self):
        return (self.extent[1] - self.extent[0]) * (self.extent[3] - self.extent[2])

    @property
    def area_per_pixel(self):
        return self.area / (self.npix_width * self.npix_height)

    @property
    def volume(self):
        return self.width * self.height * self.depth

    @property
    def volume_per_pixel(self):
        return self.volume / (self.npix_width * self.npix_height)



import numpy as np
from . import util

"""
Files to modify to implement these changes:
arepo_image
image_creators
all projectors
the slicer
arepo snapshot

TODO: Make it possible to pass angles, so that one may easily
change the viewing inclination in small steps in angle.
"""


@util.remove_astro_units
def normalize_vector(vector):
    """
    Normalize a vector and return it as a numpy array
    """
    return vector / np.linalg.norm(vector)


@util.remove_astro_units
def find_a_perpendicular_vector(vector):
    """
    Given an input vector, find a vector which is perpendicular.
    Assumes 3D vectors.
    """
    assert np.linalg.norm(vector) > 0.
    if vector[1] == 0 and vector[2] == 0:
        return np.cross(vector, [0, 1, 0])
    return np.cross(vector, [1, 0, 0])


def _assert_perpendicular(vector1, vector2, tol=1e-15):
    assert abs(np.dot(vector1, vector2)) < tol


@util.remove_astro_units
def get_basis_from_2vecs(vector1, vector2):
    """
    Returns a right-hand orthogonal basis, e1, e2, e3,
    where e1 is in the direction of vector1.
    Adapted from code by Ewald Puchwein.
    """

    # unit vector in vector1 direction
    e1 = vector1 / np.linalg.norm(vector1)

    e2 = vector2 - np.dot(e1, vector2) * e1
    # unit vector along direction of vector2-component perpendicular to vector1
    e2 = e2 / np.linalg.norm(e2)

    e3 = np.cross(e1, e2)
    assert np.fabs(np.linalg.norm(e3) - 1.0) < 1.0e-10

    # 3rd otrhogonal unit vector
    e3 = e3 / np.linalg.norm(e3)

    _assert_perpendicular(e1, e2)
    _assert_perpendicular(e2, e3)
    _assert_perpendicular(e3, e1)

    return e1, e2, e3


class Orientation:
    def __init__(self, normal_vector=None, perp_vector1=None):
        """
        Inputs:

        normal_vector: a vector which will be normal to the face
                       of the image, e.g. pass the angular momentum vector
                       of a galaxy here for a face on projection.

        perp_vector1: Perpendicular vector to the image plane, e.g.,
                         pass the angular momentum vector here for an edge-on
                         projection.

        It's possible to supply both inputs.
        """

        # Construct unit vectors for the chosen coordinate system
        self._construct_unit_vectors(normal_vector, perp_vector1)

    def _construct_unit_vectors(self, normal_vector, perp_vector1):
        """
        Construct right-handed coordinate system
        """

        if normal_vector is None and perp_vector1 is None:
            err_msg = "You have to set either normal_vector, perp_vector1 or both"
            raise RuntimeError(err_msg)

        elif normal_vector is not None and perp_vector1 is not None:
            pass
        else:
            # Normal vector given
            if normal_vector is not None:
                perp_vector1 = find_a_perpendicular_vector(normal_vector)
            # in plane vector given
            elif perp_vector1 is not None:
                normal_vector = find_a_perpendicular_vector(perp_vector1)

        # Get basis vectors
        e1, e2, e3 = get_basis_from_2vecs(normal_vector, perp_vector1)

        # Set unitless unit-vectors
        self.normal_vector = e1
        self.perp_vector1 = e2
        self.perp_vector2 = e3

    def _get_radians(self, degrees=None, radians=None):
        if radians is not None:
            return radians
        return 2.0 * np.pi * degrees / 360.

    def _get_rotation_matrix(self, axis, degrees=None, radians=None):
        """
        Standard implementation of rotation matrices around the
        x, y, and z coordinate axes. Rotation matrices for rotation
        around the coordinate axes in the right-handed
        coordinate system (perp_vector1, perp_vector2, normal_vector)
        is done by transforming this system to a coordinate
        system where these three axes are aligned along x, y and z,
        respectively, applying the either Rx, Ry or Rz and
        then transforming back.
        """

        radians = self._get_radians(degrees)

        if axis == 'x':
            R = np.array([[1, 0, 0],
                         [0, np.cos(radians), -np.sin(radians)],
                         [0, np.sin(radians), np.cos(radians)]])
            return R
        elif axis == 'y':
            R = np.array([[np.cos(radians), 0, np.sin(radians)],
                         [0, 1, 0],
                         [-np.sin(radians), 0, np.cos(radians)]])
            return R
        elif axis == 'z':
            R = np.array([[np.cos(radians), -np.sin(radians), 0],
                         [np.sin(radians), np.cos(radians), 0],
                         [0, 0, 1]])
            return R
        elif axis == 'perp_vector1':
            u = self.perp_vector1
        elif axis == 'perp_vector2':
            u = self.perp_vector2
        elif axis == 'normal_vector':
            u = self.normal_vector
        #     Rx = self._get_rotation_matrix('x', degrees, radians)
        #     Rtmp = np.matmul(Rx, self.inverse_rotation_matrix)
        #     R = np.matmul(self.rotation_matrix, Rtmp)

        # elif axis == 'perp_vector2':
        #     Ry = self._get_rotation_matrix('y', degrees, radians)
        #     Rtmp = np.matmul(Ry, self.inverse_rotation_matrix)
        #     R = np.matmul(self.rotation_matrix, Rtmp)

        # elif axis == 'normal_vector':
        #     Rz = self._get_rotation_matrix('z', degrees, radians)
        #     Rtmp = np.matmul(Rz, self.inverse_rotation_matrix)
        #     R = np.matmul(self.rotation_matrix, Rtmp)

        else:
            raise RuntimeError("invalid input")

        # Construct rotation matrix that rotates radians around
        # unit vector 'u'
        # see e.g. https://en.wikipedia.org/w/index.php?title=Rotation_matrix
        u_cross = np.array([[0, -u[2], u[1]],
                            [u[2], 0, -u[0]],
                            [-u[1], u[0], 0]])
        R = np.cos(radians) * np.identity(3) + np.sin(radians) * u_cross \
            + (1 - np.cos(radians)) * np.outer(u, u)

        return R

    def _apply_rotation_matrix(self, R):
        self.normal_vector = np.matmul(R, self.normal_vector)
        self.perp_vector1 = np.matmul(R, self.perp_vector1)
        self.perp_vector2 = np.matmul(R, self.perp_vector2)

    def rotate_around_x(self, degrees=None, radians=None):
        """
        """
        R = self._get_rotation_matrix('x', degrees, radians)
        self._apply_rotation_matrix(R)

    def rotate_around_y(self, degrees=None, radians=None):
        """
        """
        R = self._get_rotation_matrix('y', degrees, radians)
        self._apply_rotation_matrix(R)

    def rotate_around_z(self, degrees=None, radians=None):
        """
        """
        R = self._get_rotation_matrix('z', degrees, radians)
        self._apply_rotation_matrix(R)

    def rotate_around_normal_vector(self, degrees=None, radians=None):
        """
        """
        R = self._get_rotation_matrix('normal_vector', degrees, radians)
        self._apply_rotation_matrix(R)

    def rotate_around_perp_vector1(self, degrees=None, radians=None):
        """
        """
        R = self._get_rotation_matrix('perp_vector1', degrees, radians)
        self._apply_rotation_matrix(R)

    def rotate_around_perp_vector2(self, degrees=None, radians=None):
        """
        """
        R = self._get_rotation_matrix('perp_vector2', degrees, radians)
        self._apply_rotation_matrix(R)

    @property
    def inverse_rotation_matrix(self):
        """
        An inverse rotation matrix, which rotates from the Orientation
        system into a standard xyz coordinate system. We have the
        following properties:

        np.matmul(inverse_rotation_matrix, perp_vector1) = [1, 0, 0]
        np.matmul(inverse_rotation_matrix, perp_vector2) = [0, 1, 0]
        np.matmul(inverse_rotation_matrix, normal_vector) = [0, 0, 1]

        The use case for the inverse rotation matrix is to transform
        Arepo data into a coordinate system where the perp_vector1 direction
        is in the x-direction, the perp_vector2 direction is in the y-direction
        and the normal_vector is in the z-direction.
        """
        ex = self.cartesian_unit_vectors['x']
        ey = self.cartesian_unit_vectors['y']
        ez = self.cartesian_unit_vectors['z']
        return np.array([ex, ey, ez])

    @property
    def rotation_matrix(self):
        """
        This rotation matrix rotates the standard cartesian unit vectors

        ex = [1, 0, 0]
        ey = [0, 1, 0]
        ez = [0, 0, 1]

        into the cartesian coordinate system defined by the three unit vectors

        perp_vector1
        perp_vector2
        normal_vector

        which can also be accessed through the cartesian_unit_vectors method.

        That is, the following equality holds:

        perp_vector1 = np.matmul(rotation_matrix, [1, 0, 0])
        perp_vector2 = np.matmul(rotation_matrix, [0, 1, 0])
        normal_vector = np.matmul(rotation_matrix, [0, 0, 1])
        """
        return self.inverse_rotation_matrix.T

    @property
    def cartesian_unit_vectors(self):
        """
        The x, y, and z unit vectors in the Orientation
        given in terms of their components in the original simulation
        coordinate system.
        """
        unit_vectors = {}
        unit_vectors['x'] = self.perp_vector1
        unit_vectors['y'] = self.perp_vector2
        unit_vectors['z'] = self.normal_vector
        return unit_vectors

    @property
    def spherical_unit_vectors(self):
        raise RuntimeError('not implemented')

    @property
    def cylindrical_unit_vectors(self):
        raise RuntimeError('not implemented')

    @property
    def euler_angles(self):
        # TODO: intrinsic and extrinsic version?
        raise RuntimeError('not implemented')

    @property
    def copy(self):
        """
        Return a copy of the current Orientation instance.
        """
        return Orientation(normal_vector=self.normal_vector,
                           perp_vector1=self.perp_vector1)

    def _are_equal(self, orientation2):
        """
        Compare the current orientation instance with a
        different one. Returns True if they are identical
        and False if not.
        """
        u1 = self.cartesian_unit_vectors
        u2 = orientation2.cartesian_unit_vectors
        identical = True
        for key in u1:
            identical = np.allclose(u1[key], u2[key], atol=1e-15) and identical
        return identical

    def __print__(self):
        print('normal_vector:', self.normal_vector)
        print('perp_vector1: ', self.perp_vector1)
        print('perp_vector2: ', self.perp_vector2)

    def __repr__(self):
        s = 'Paicos Orientation instance\n'
        s += f'normal_vector: {self.normal_vector}\n'
        s += f'perp_vector1:  {self.perp_vector1}\n'
        s += f'perp_vector2:  {self.perp_vector2}'
        return s



import numba
import math
import numpy as np
from .bvh_tree import leading_zeros_cython as count_leading_zeros

# Hardcoded for uint64 coordinates (don't change without also changing morton
# code functions)
L = 21


def leading_zeros_python(key64bit):
    """
    1 us per calculation, so super slow
    compared to the 3 ns per calculation that one gets
    with Cython.
    """
    return format(key64bit, '064b').find('1')


@numba.jit(nopython=True)
def part1by2_64(n):
    n &= 0x1fffff
    n = (n | (n << 32)) & 0x1f00000000ffff
    n = (n | (n << 16)) & 0x1f0000ff0000ff
    n = (n | (n << 8)) & 0x100f00f00f00f00f
    n = (n | (n << 4)) & 0x10c30c30c30c30c3
    n = (n | (n << 2)) & 0x1249249249249249
    return n


@numba.jit(nopython=True)
def unpart1by2_64(n):
    n = numba.uint64(n)
    n &= 0x1249249249249249
    n = (n ^ (n >> 2)) & 0x10c30c30c30c30c3
    n = (n ^ (n >> 4)) & 0x100f00f00f00f00f
    n = (n ^ (n >> 8)) & 0x1f0000ff0000ff
    n = (n ^ (n >> 16)) & 0x1f00000000ffff
    n = (n ^ (n >> 32)) & 0x1fffff
    return n


@numba.jit(nopython=True)
def encode64(x, y, z):
    xx = part1by2_64(x)
    yy = part1by2_64(y)
    zz = part1by2_64(z)
    return xx * 4 + yy * 2 + zz


@numba.jit(nopython=True)
def decode64(key):
    x = unpart1by2_64(key >> 2)
    y = unpart1by2_64(key >> 1)
    z = unpart1by2_64(key)
    return numba.float64(x), numba.float64(y), numba.float64(z)


@numba.jit(nopython=True)
def get_morton_keys64(pos):
    morton_keys = np.empty(pos.shape[0], dtype=np.uint64)
    for ip in range(pos.shape[0]):
        morton_keys[ip] = encode64(pos[ip, 0], pos[ip, 1], pos[ip, 2])
    return morton_keys


# @numba.jit(nopython=True, inline='always')
def findSplit(sortedMortonCodes, first, last):
    # Identical Morton codes => split the range in the middle.

    firstCode = sortedMortonCodes[first]
    lastCode = sortedMortonCodes[last]

    if (firstCode == lastCode):
        return (first + last) >> 1

    # Calculate the number of highest bits that are the same
    # for all objects, using the count-leading-zeros intrinsic.

    commonPrefix = count_leading_zeros(firstCode ^ lastCode)

    # Use binary search to find where the next bit differs.
    # Specifically, we are looking for the highest object that
    # shares more than commonPrefix bits with the first one.

    split = first  # initial guess
    step = last - first

    while (step > 1):
        step = (step + 1) >> 1  # exponential decrease
        newSplit = split + step  # proposed new position

        if (newSplit < last):
            splitCode = sortedMortonCodes[newSplit]
            splitPrefix = count_leading_zeros(firstCode ^ splitCode)
            if (splitPrefix > commonPrefix):
                split = newSplit  # accept proposal

    return split


# @numba.jit(nopython=True)
def generateHierarchy(sortedMortonCodes, tree_children, tree_parents):
    """
    Generate tree using the sorted Morton code.
    This is done in parallel, see this blogpost
    https://developer.nvidia.com/blog/thinking-parallel-part-iii-tree-construction-gpu/
    and the corresponding paper:

    """
    n_codes = sortedMortonCodes.shape[0]
    num_internal_nodes = n_codes - 1

    # Launch thread for each internal node
    for idx in range(num_internal_nodes):

        # Find out which range of objects the node corresponds to.
        # (This is where the magic happens!)

        # Determine range
        first, last = determineRange(sortedMortonCodes, n_codes, idx)

        # Determine where to split the range.
        split = findSplit(sortedMortonCodes, first, last)

        # Select childA.
        if (split == first):
            # Is a leaf
            childA = split + num_internal_nodes
        else:
            # Is in an internal node
            childA = split

        # Select childB.
        if (split + 1 == last):
            childB = split + 1 + num_internal_nodes
        else:
            childB = split + 1

        # Record parent-child relationships.
        tree_children[idx, 0] = childA
        tree_children[idx, 1] = childB
        tree_parents[childA] = idx
        tree_parents[childB] = idx


# @numba.jit(nopython=True, inline='always')
def determineRange(sortedMortonCodes, n_codes, idx):
    """
    The determine range function needed by the generateHierarchy function.
    This code has has been adapted from the CornerStone CUDA/C++ code,
    see constructInternalNode in btree.hpp.
    """
    d = 1
    minPrefixLength = -1
    firstIndex = idx
    max_last = n_codes

    if (firstIndex > 0):
        p1 = sortedMortonCodes[firstIndex] ^ sortedMortonCodes[firstIndex + 1]
        m1 = sortedMortonCodes[firstIndex] ^ sortedMortonCodes[firstIndex - 1]

        # Count leading zeros
        lz_p1 = count_leading_zeros(p1)
        lz_m1 = count_leading_zeros(m1)

        if lz_p1 > lz_m1:
            d = 1
            minPrefixLength = lz_m1
        else:
            d = -1
            minPrefixLength = lz_p1

    searchRange = 2
    secondIndex = firstIndex + searchRange * d

    while (0 <= secondIndex and secondIndex < max_last):
        lz_first_second = count_leading_zeros(
            sortedMortonCodes[firstIndex] ^ sortedMortonCodes[secondIndex])
        if lz_first_second > minPrefixLength:
            searchRange *= 2
            secondIndex = firstIndex + searchRange * d
            if secondIndex < max_last:
                lz_first_second = count_leading_zeros(
                    sortedMortonCodes[firstIndex] ^ sortedMortonCodes[secondIndex])
            else:
                break
        else:
            break

    # start binary search with known searchRange
    secondIndex = firstIndex
    while searchRange > 1:
        searchRange = (searchRange + 1) // 2
        newJdx = secondIndex + searchRange * d
        if (0 <= newJdx and newJdx < max_last):
            lz_f_Jdx = count_leading_zeros(
                sortedMortonCodes[firstIndex] ^ sortedMortonCodes[newJdx])
            if lz_f_Jdx > minPrefixLength:
                secondIndex = newJdx

    first = min(secondIndex, firstIndex)
    last = max(secondIndex, firstIndex)

    return first, last


def find_bounding_volume_as_sfc_keys(center, hsml):
    min_val_x = numba.uint64(max(center[0] - hsml, 0))
    min_val_y = numba.uint64(max(center[1] - hsml, 0))
    min_val_z = numba.uint64(max(center[2] - hsml, 0))
    max_val_x = numba.uint64(min(math.ceil(center[0] + hsml), 2**L - 1))
    max_val_y = numba.uint64(min(math.ceil(center[1] + hsml), 2**L - 1))
    max_val_z = numba.uint64(min(math.ceil(center[2] + hsml), 2**L - 1))
    sfc_key_min = encode64(min_val_x, min_val_y, min_val_z)
    sfc_key_max = encode64(max_val_x, max_val_y, max_val_z)
    return sfc_key_min, sfc_key_max


def set_leaf_bounding_volumes(tree_bounds, points, half_size, conversion_factor):
    num_internal_nodes = half_size.shape[0] - 1
    num_leafs = half_size.shape[0]
    f = conversion_factor
    to_int = numba.uint64
    for ip in range(num_leafs):
        center = points[ip, :]
        x_min = to_int(max((center[0] - f * half_size[ip]), 0))
        y_min = to_int(max((center[1] - f * half_size[ip]), 0))
        z_min = to_int(max((center[2] - f * half_size[ip]), 0))
        x_max = to_int(
            min(math.ceil((center[0] + f * half_size[ip])), 2**L - 1))
        y_max = to_int(
            min(math.ceil((center[1] + f * half_size[ip])), 2**L - 1))
        z_max = to_int(
            min(math.ceil((center[2] + f * half_size[ip])), 2**L - 1))
        tree_bounds[ip + num_internal_nodes, 0, 0] = x_min
        tree_bounds[ip + num_internal_nodes, 1, 0] = y_min
        tree_bounds[ip + num_internal_nodes, 2, 0] = z_min
        tree_bounds[ip + num_internal_nodes, 0, 1] = x_max
        tree_bounds[ip + num_internal_nodes, 1, 1] = y_max
        tree_bounds[ip + num_internal_nodes, 2, 1] = z_max


def propagate_bounds_upwards(tree_bounds, tree_parents, tree_children):
    num_internal_nodes = tree_children.shape[0]
    num_leafs = num_internal_nodes + 1

    next_parents = []
    for ip in range(num_leafs):
        next_parents.append(tree_parents[ip + num_internal_nodes])
    next_parents = np.unique(next_parents)

    box = np.zeros((3, 2), dtype=np.uint64)
    while len(next_parents) > 0:
        current_parents = np.array(next_parents)
        next_parents = []
        for ii in range(current_parents.shape[0]):
            node_id = current_parents[ii]
            next_parent = tree_parents[node_id]
            if next_parent != -1:
                next_parents.append(tree_parents[node_id])
            childA = tree_children[node_id, 0]
            childB = tree_children[node_id, 1]
            boundsA = tree_bounds[childA, :, :]
            boundsB = tree_bounds[childB, :, :]
            box[0, 0] = np.min([boundsB[0, 0], boundsA[0, 0]])
            box[1, 0] = np.min([boundsB[1, 0], boundsA[1, 0]])
            box[2, 0] = np.min([boundsB[2, 0], boundsA[2, 0]])
            box[0, 1] = np.max([boundsB[0, 1], boundsA[0, 1]])
            box[1, 1] = np.max([boundsB[1, 1], boundsA[1, 1]])
            box[2, 1] = np.max([boundsB[2, 1], boundsA[2, 1]])

            tree_bounds[node_id, :, :] = box[:, :]
        next_parents = np.unique(next_parents)
        # if len(next_parents) < 10:
        #     print(node_id, len(next_parents), next_parents)


@numba.jit(nopython=True)
def distance(xf_q, yf_q, zf_q, xf, yf, zf):
    dist = math.sqrt((xf - xf_q)**2
                     + (yf - yf_q)**2
                     + (zf - zf_q)**2)
    return dist


# @numba.jit(nopython=True)
# def is_within_bounds(xf_q, yf_q, zf_q, tree_bound):
#     x_min, y_min, z_min = tree_bound[:, ]decode64(tree_bound[0])
#     x_max, y_max, z_max = decode64(tree_bound[1])
#     if x_min > xf_q or x_max < xf_q:
#         return False
#     if y_min > yf_q or y_max < yf_q:
#         return False
#     if z_min > zf_q or z_max < zf_q:
#         return False
#     return True

@numba.jit(nopython=True)
def is_point_in_box(point, box):
    for ii in range(3):
        if point[ii] < box[ii, 0] or point[ii] > box[ii, 1]:
            return False
    return True


@numba.jit(nopython=True)
def find_nearest_neighbors(points, tree_parents, tree_children, tree_bounds,
                           query_points, dists, ids, start_id=0):
    n_queries = query_points.shape[0]
    for ip in range(n_queries):
        num_internal_nodes = numba.int64(tree_children.shape[0])

        # Select a query_point from the list of queries
        query_point = query_points[ip]

        if start_id != 0:
            this_query_start_id = int(start_id)
            # print(this_query_start_id, start_id)
            is_leaf = start_id >= num_internal_nodes
            in_box = is_point_in_box(
                query_point, tree_bounds[this_query_start_id])
            while not in_box or is_leaf:
                # print(this_query_start_id, start_id)
                this_query_start_id = tree_parents[this_query_start_id]
                # print('dfdf')
                in_box = is_point_in_box(
                    query_point, tree_bounds[this_query_start_id])
                # print('22222')
                is_leaf = this_query_start_id >= num_internal_nodes
                # print('33333', is_leaf, in_box)
                # print(this_query_start_id, start_id)
                if this_query_start_id == -1:
                    # print('went all the way up...')
                    this_query_start_id = 0
                    break
        else:
            this_query_start_id = 0

        # print('did we go here?')
        # print('this_query_start_id', this_query_start_id, start_id)
        # We traverse the nodes and leafs using a while loop and a queue.

        # Local memory on each tread (32 should be fine?)
        queue = np.zeros(256, dtype=np.int64)
        # Initialize queue_index an start at node 0
        queue_index = 0
        queue[queue_index] = this_query_start_id

        # Initialize min_dist, and min_index
        min_dist = (2**L - 1.0)
        min_index = -1

        while queue_index >= 0:

            node_id = queue[queue_index]

            childA = tree_children[node_id, 0]
            childB = tree_children[node_id, 1]

            is_leafA = childA >= num_internal_nodes
            is_leafB = childB >= num_internal_nodes

            point_in_A = is_point_in_box(query_point, tree_bounds[childA])
            point_in_B = is_point_in_box(query_point, tree_bounds[childB])

            # Do explicit check if in a leaf
            # print(f'node_id: {node_id}\t childA: {childA}\t childB: {childB}\t')
            # if node_id == 6:
            #     print('node_id, point_in_A, point_in_B, is_leaf,
            # is_leafB', node_id, point_in_A, point_in_B, is_leafA, is_leafB)
            #     print(query_point)
            #     print(tree_bounds[childA])
            #     print(tree_bounds[childB])
            # raise RuntimeError("d")
            if point_in_A and is_leafA:
                data_id = childA - num_internal_nodes
                dist = distance(query_point[0],
                                query_point[1],
                                query_point[2],
                                points[data_id, 0],
                                points[data_id, 1],
                                points[data_id, 2])

                # if node_id == 6:
                #     print(f'dist: {dist}, min_dist: {min_dist}')
                #     print(dist, min_dist, data_id)
                #     print(points[data_id], query_point)
                # raise RuntimeError("d")
                if dist < min_dist:
                    min_dist = dist
                    min_index = data_id

            if point_in_B and is_leafB:
                data_id = childB - num_internal_nodes
                dist = distance(query_point[0],
                                query_point[1],
                                query_point[2],
                                points[data_id, 0],
                                points[data_id, 1],
                                points[data_id, 2])

                if dist < min_dist:
                    min_dist = dist
                    min_index = data_id

            # Whether to traverse
            traverseA = point_in_A and not is_leafA
            traverseB = point_in_B and not is_leafB

            if (not traverseA) and (not traverseB):
                queue_index -= 1
            else:
                if traverseA:
                    queue[queue_index] = childA
                else:
                    queue[queue_index] = childB
                if traverseA and traverseB:
                    queue_index += 1
                    queue[queue_index] = childB

        # Outside main while loop
        dists[ip] = min_dist
        ids[ip] = min_index  # + num_internal_nodes + 1
    return dists, ids


class BinaryTree:
    def __init__(self, positions, sizes):
        """
        Python/Numba implementation of a BVH (boundary volume hierarchy) tree.
        """

        # Copy array
        self._pos = np.array(positions)

        # Convert self._position to units from [0, 1) * 2**L
        self.off_sets = np.min(self._pos, axis=0)
        self._pos -= self.off_sets[None, :]

        max_pos = np.max(self._pos)
        self.conversion_factor = (2**L - 1) / max_pos
        self._pos *= self.conversion_factor

        # if use_units take the value...
        # self.pos = self.pos.value
        # self.

        # Calculate uint64 and Morton keys
        self._pos_uint = self._pos.astype(np.uint64)
        self.morton_keys = get_morton_keys64(self._pos_uint)

        # Store index for going back to original ids.
        self.sort_index = np.argsort(self.morton_keys)

        # Do the sorting
        self.morton_keys = self.morton_keys[self.sort_index]
        self._pos = self._pos[self.sort_index, :]
        self._pos_uint = self._pos_uint[self.sort_index, :]

        # Some tree properties
        self.num_leafs = self.morton_keys.shape[0]
        self.num_internal_nodes = self.num_leafs - 1
        self.num_leafs_and_nodes = 2 * self.num_leafs - 1

        # Allocate arrays
        self.children = -1 * np.ones((self.num_internal_nodes, 2), dtype=int)
        self.parents = -1 * np.ones(self.num_leafs_and_nodes, dtype=int)

        # This sets the parent and children properties
        generateHierarchy(self.morton_keys, self.children, self.parents)

        # Set the boundaries for the leafs
        self.bounds = np.zeros((self.num_leafs_and_nodes, 3, 2), dtype=np.uint64)

        set_leaf_bounding_volumes(self.bounds, self._pos,
                                  sizes[self.sort_index],
                                  self.conversion_factor)

        # Calculate the bounding volumes for internal nodes,
        # by propagating the information upwards in the tree
        propagate_bounds_upwards(self.bounds, self.parents, self.children)

    def _to_tree_coordinates(self, pos):
        return (pos - self.off_sets[None, :]) * self.conversion_factor

    def _tree_node_ids_to_data_ids(self, ids):
        """
        Go from sorted ids to original ordering of
        *selected* data.
        """
        return np.arange(self.sort_index.shape[0])[self.sort_index][ids]

    def nearest_neighbor(self, query_points, start_id=0):
        if query_points.ndim == 1:
            n_queries = 1
        else:
            n_queries = query_points.shape[0]

        ids = np.zeros(n_queries, dtype=np.int64)
        dists = np.zeros(n_queries, dtype=np.float64)
        find_nearest_neighbors(self._pos, self.parents, self.children, self.bounds,
                               self._to_tree_coordinates(
                                   query_points), dists, ids,
                               start_id=start_id)
        return dists / self.conversion_factor, self._tree_node_ids_to_data_ids(ids)



"""
Defines small_a and small_h, conversion to Gauss and eV,
and the PaicosQuantity and PaicosTimeSeries classes.
"""
from fractions import Fraction
import numpy as np
import astropy.units as u
from astropy.units import Quantity
from astropy.units import UnitConversionError

_ns = globals()

small_a = u.def_unit(
    ["small_a"],
    prefixes=False,
    namespace=_ns,
    doc="Cosmological scale factor.",
    format={"latex": "a", "unicode": r"a"},
)
u.def_physical_type(small_a, "scale factor")


small_h = u.def_unit(
    ["small_h"],
    namespace=_ns,
    prefixes=False,
    doc='Reduced/"dimensionless" Hubble constant',
    format={"latex": r"h", "unicode": r"h"},
)

u.add_enabled_units(small_a)
u.add_enabled_units(small_h)

# Allows conversion between K and eV
u.add_enabled_equivalencies(u.temperature_energy())
# Allows conversion to Gauss (potential issues?)
# https://github.com/astropy/astropy/issues/7396

# The equivalencies sometimes frustratingly fail.
# TODO: Override the .to method ignore the Gauss component of the unit...
gauss_B = (u.g / u.cm)**(0.5) / u.s
equiv_B = [(u.G, gauss_B, lambda x: x, lambda x: x)]
Bscaling = small_a**(-2) * small_h
equiv_B_comoving = [(u.G * Bscaling, gauss_B * Bscaling, lambda x: x, lambda x: x)]
Bscaling = small_a**(-2)
equiv_B_no_small_h = [(u.G * Bscaling, gauss_B * Bscaling, lambda x: x, lambda x: x)]
u.add_enabled_equivalencies(equiv_B)
u.add_enabled_equivalencies(equiv_B_comoving)
u.add_enabled_equivalencies(equiv_B_no_small_h)


def get_unit_dictionaries(unit):
    """
    Returns dictionaries with information about the units of the
    quantity.
    """
    codic = {}
    dic = {}
    for base, power in zip(unit.bases, unit.powers):
        if base in (small_a, small_h):
            codic[base] = power
        else:
            dic[base] = power
    for key in [small_h, small_a]:
        if key not in codic:
            codic[key] = 0

    return codic, dic


def construct_unit_from_dic(dic):
    """
    Construct unit from a dictionary with the format returned
    from __get_unit_dictionaries
    """
    return np.prod([unit**dic[unit] for unit in dic])


def separate_units(unit):
    """
    Separate the standard physical units (u_unit) from the units involving
    a and h (pu_unit). That is,

    u_unit, pu_unit = separate_units(unit)

    where pu_unit contains small_a and small_h and u_unit contains
    everything else such that unit = u_unit * pu_unit
    """
    codic, dic = get_unit_dictionaries(unit)
    u_unit = construct_unit_from_dic(dic)
    pu_unit = construct_unit_from_dic(codic)
    return u_unit, pu_unit


def get_new_unit(unit, remove_list=[]):
    """
    Return new unit where base units in the remove list have been removed.
    """
    unit_list = []
    for base, power in zip(unit.bases, unit.powers):
        if base not in remove_list:
            unit_list.append(base**power)

    return np.prod(unit_list)


class PaicosQuantity(Quantity):

    """
    PaicosQuantity is a subclass of the astropy Quantity class which
    represents a number with some associated unit.

    This subclass in addition includes a and h factors used in the definition
    of comoving variables.

    Parameters
    ----------

    value: the numeric values of your data (similar to astropy Quantity)

    a: the cosmological scale factor of your data

    h: the reduced Hubble parameter, e.g. h = 0.7

    unit: a string, e.g. 'g/cm^3 small_a^-3 small_h^2' or astropy Unit
    The latter can be defined like this:

    from paicos import units as pu
    from astropy import units as u
    unit = u.g*u.cm**(-3)*small_a**(-3)*small_h**(2)

    The naming of small_a and small_h is to avoid conflict with the already
    existing 'annum' (i.e. a year) and 'h' (hour) units.

    Returns
    ----------

    Methods/properties
    ----------

    no_small_h: returns a new comoving quantity where the h-factors have
               been removed and the numeric value adjusted accordingly.

    to_physical: returns a new  object where both a and h factors have been
                 removed, i.e. we have switched from comoving values to
                 the physical value.

    label: Return a Latex label for use in plots.

    Examples
    ----------

    units = 'g cm^-3 small_a^-3 small_h^2'
    A = PaicosQuantity(2, units, h=0.7, a=1/128)

    # Create a new comoving quantity where the h-factors have been removed
    B = A.no_small_h

    # Create a new quantity where both a and h factor have been removed,
    # i.e. we have switched from a comoving quantity to the physical value

    C = A.to_physical

    """

    # pylint: disable=too-many-arguments

    def __new__(cls, value, unit=None, dtype=None, copy=False, order=None,
                subok=False, ndmin=0, h=None, a=None, comoving_sim=None):
        """
        Here we initialize the Paicos Quantity. The three additional
        input arguments (compared to a standard astropy quantity) are

        h
        a
        comoving_sim
        """

        assert h is not None, 'Paicos quantity is missing a value for h'
        assert a is not None, 'Paicos quantity is missing a value for a'
        assert comoving_sim is not None, 'is this from a comoving_sim?'

        if hasattr(value, 'unit'):
            unit = value.unit
            value = value.value

        obj = super().__new__(cls, value, unit=unit, dtype=dtype, copy=copy,
                              order=order, subok=subok, ndmin=ndmin)

        obj._h = h
        obj._a = a
        obj._comoving_sim = comoving_sim

        return obj

    def __array_finalize__(self, obj):
        """
        Heavily inspired by the astropy Quantity version
        """
        super_array_finalize = super().__array_finalize__
        if super_array_finalize is not None:
            super_array_finalize(obj)

        # If we're a new object or viewing an ndarray, nothing has to be done.
        if obj is None or obj.__class__ is np.ndarray:
            return

        # Set Paicos specific parameters
        self._h = getattr(obj, '_h', None)
        self._a = getattr(obj, '_a', None)
        self._comoving_sim = getattr(obj, '_comoving_sim', None)

    @property
    def a(self):
        """
        The scale factor.
        """
        if self.comoving_sim:
            return self._a
        raise RuntimeError('Non-comoving object has no scale factor')

    @property
    def h(self):
        """
        The reduced Hubble parameter
        """
        return self._h

    @property
    def comoving_sim(self):
        """
        Whether the simulation had ComovingIntegrationOn
        """
        return self._comoving_sim

    @property
    def z(self):
        """
        The redshift.
        """
        if self.comoving_sim:
            return 1. / self._a - 1.
        raise RuntimeError('Non-comoving object has no redshift')

    def lookback_time(self, reader_object):
        """
        The lookback time.

        Requires a reader object as input (e.g. a snap,
        cat or other instance of Snapshot, Catalog or PaicosReader)
        """
        if self.comoving_sim:
            return reader_object.get_lookback_time(self.z)
        msg = 'lookback_time not defined for non-comoving sim'
        raise RuntimeError(msg)

    def age(self, reader_object):
        """
        The age of the universe in the simulation.

        Requires a reader object as input (e.g. a snap,
        cat or other instance of Snapshot, Catalog or PaicosReader)
        """
        if self.comoving_sim:
            return reader_object.get_age(self.z)
        msg = 'age not defined for non-comoving sim'
        raise RuntimeError(msg)

    @property
    def time(self):
        """
        The time elapsed since the beginning of the simulation.

        Only defined for non-comoving simulations.
        """
        if self.comoving_sim:
            msg = 'time not defined for comoving sim'
            raise RuntimeError(msg)

        if hasattr(self._a, 'unit'):
            return self._a
        else:
            return self._a * u.Unit('arepo_time')

    @property
    def unit_quantity(self):
        """
        Returns a new PaicosQuantity with the same units as the current
        PaicosQuantity and a numeric value of 1.
        """
        return PaicosQuantity(1., self.unit, a=self._a, h=self.h,
                              comoving_sim=self.comoving_sim)

    @property
    def copy(self):
        """
        Returns a copy of the PaicosQuantity
        """
        return PaicosQuantity(np.array(self.value), self.unit, a=self._a, h=self.h,
                              comoving_sim=self.comoving_sim, copy=True)

    @property
    def uq(self):
        """
        A short hand for the 'unit_quantity' method.
        """
        return self.unit_quantity

    @property
    def hdf5_attrs(self):
        """
        Give the units as a dictionary for hdf5 data set attributes
        """
        return {'unit': self.unit.to_string()}

    @property
    def no_small_h(self):
        """
        Remove scaling with h, returning a quantity with adjusted values.
        """
        codic, _ = get_unit_dictionaries(self.unit)
        factor = self.h**codic[small_h]

        value = self.view(np.ndarray)
        new_unit = get_new_unit(self.unit, [small_h])
        return self._new_view(value * factor, new_unit)

    @property
    def cgs(self):
        """
        Returns a copy of the current `PaicosQuantity` instance with CGS units.
        The value of the resulting object will be scaled.
        """
        u_unit, pu_unit = separate_units(self.unit)
        cgs_unit = u_unit.cgs
        new_unit = pu_unit * cgs_unit / cgs_unit.scale
        return self._new_view(self.value * cgs_unit.scale, new_unit)

    @property
    def si(self):
        """
        Returns a copy of the current `PaicosQuantity` instance with SI units.
        The value of the resulting object will be scaled.
        """
        u_unit, pu_unit = separate_units(self.unit)
        si_unit = u_unit.si
        new_unit = pu_unit * si_unit / si_unit.scale
        return self._new_view(self.value * si_unit.scale, new_unit)

    def to(self, unit, equivalencies=[], copy=True):
        """
        Convert to different units. Similar functionality to the astropy
        Quantity.to() method.
        """
        if isinstance(unit, str):
            unit = u.Unit(unit)

        u_unit, pu_unit = separate_units(self.unit)

        u_unit_to, pu_unit_to = separate_units(unit)

        if pu_unit_to == u.Unit(''):
            return super().to(unit * pu_unit, equivalencies, copy)
        elif pu_unit == pu_unit_to:
            return super().to(unit, equivalencies, copy)
        else:
            err_msg = ('\n\nYou have requested conversion from\n {} '
                       + '\nto\n {} . \nThis is not possible as the a and h '
                       + 'factors differ. I.e. you cannot convert from\n '
                       + '{}\nto\n {}\nUse the .to_physical or .no_small_h '
                       + 'methods if you are trying to get rid of the a '
                       + 'and h factors.')

            raise RuntimeError(err_msg.format(self.unit, unit, pu_unit, pu_unit_to))

    @property
    def arepo(self):
        """
        Returns the quantity in Arepo code units.
        """
        arepo_bases = set([u.Unit('arepo_mass'),
                           u.Unit('arepo_length'),
                           u.Unit('arepo_time')])
        try:
            return self.decompose(bases=arepo_bases)
        except UnitConversionError as inst:
            err_msg = ('Conversion to arepo_units does not work well for '
                       + 'Temperature and magnetic field strength in Gauss. '
                       + 'Astropy throws the following error: ' + str(inst))

            raise UnitConversionError(err_msg) from inst

        return None

    @property
    def astro(self):
        """
        Returns the quantity in typical units used in cosmological simulations
        """
        return self.decompose(bases=[u.kpc, u.Msun, u.s, u.uG, u.keV, u.K])

    def decompose(self, bases=[]):
        """
        Decompose into a different set of units, e.g.

        A = B.decompose(bases=[u.kpc, u.Msun, u.s, u.uG, u.keV])

        small_a and small_h are automatically included in the bases.
        """
        _, pu_unit = separate_units(self.unit)
        if len(bases) == 0 or pu_unit == u.Unit(''):
            return super().decompose(bases)

        if isinstance(bases, set):
            bases = list(bases)
        bases.append(small_a)
        bases.append(small_h)
        bases = set(bases)
        return super().decompose(bases)

    def label(self, variable=''):
        """
        Return a Latex string for use in plots. The optional
        input variable could be the Latex symbol for the physical variable,
        for instance \rho or \nabla\times\vec{v}.
        """

        a_sc, a_sc_str = self.__scaling_and_scaling_str(small_a)
        h_sc, h_sc_str = self.__scaling_and_scaling_str(small_h)

        co_label = a_sc_str + h_sc_str

        normal_unit = get_new_unit(self.unit, [small_h, small_a])

        _, dic = get_unit_dictionaries(self.unit)

        unit_label = ''
        for ii, key in enumerate(dic.keys()):
            _, unit_str = self.__scaling_and_scaling_str(key)
            unit_label += unit_str
            if ii < len(dic.keys()) - 1:
                unit_label += r'\;'

        label = co_label + r'\; \left[' + unit_label + r'\right]'

        # Get ckpc, cMpc, ckpc/h and Mkpc/h as used in literature
        if normal_unit in ('kpc', 'Mpc', 'Gpc'):
            if a_sc in (0, 1):
                if h_sc in (0, -1):
                    if a_sc == 1:
                        label = r'\mathrm{c}' + unit_label
                    elif a_sc == 0:
                        label = unit_label
                    if h_sc == -1:
                        label = label + r'\;h^{-1}'

            label = '[' + label + ']'

        if len(variable) > 0:
            label = variable + r'\;' + label
        label = '$' + label + '$'

        return label

    @property
    def to_physical(self):
        """
        Returns a copy of the current `PaicosQuantity` instance with the
        a and h factors removed, i.e. transform from comoving to physical.
        The value of the resulting object is scaled accordingly.
        """
        codic, _ = get_unit_dictionaries(self.unit)
        factor = self.h**codic[small_h]

        if self.comoving_sim:
            factor *= self.a**codic[small_a]

        value = self.view(np.ndarray)
        new_unit = get_new_unit(self.unit, [small_a, small_h])
        return self._new_view(value * factor, new_unit)

    def to_comoving(self, unit):
        """
        Returns a copy of the current `PaicosQuantity` instance with the
        a and h factors given by the input unit.
        """

        if not self.comoving_sim:
            raise RuntimeError('Only implemented for comoving simulations')

        if isinstance(unit, str):
            unit = u.Unit(unit)

        u_unit_to, pu_unit_to = separate_units(unit)

        u_unit, pu_unit = separate_units(self.unit)

        if u_unit_to == u.Unit(''):
            u_unit_to = u_unit
        elif u_unit_to != u_unit:
            raise RuntimeError('This method can only change the a and h factors!')

        change_pu = pu_unit / pu_unit_to

        codic, _ = get_unit_dictionaries(change_pu)
        factor = self.h**codic[small_h]

        if self.comoving_sim:
            factor *= self.a**codic[small_a]

        value = self.view(np.ndarray)
        new_unit = self.unit / change_pu

        return self._new_view(value * factor, new_unit)

    def to_any(self, new_unit):
        """
        Convert quantity to any (possible) unit.

        The .to method only allows converting standard
        physical units while the .to_comoving only allows
        modifying the a and h factors. This method allows
        changing the units in one go.
        """
        u_unit, pu_unit = separate_units(new_unit)
        new_quant = self.to(u_unit)
        new_quant = new_quant.to_comoving(pu_unit)
        return new_quant

    def __scaling_and_scaling_str(self, unit):
        """
        Helper function to create labels
        """
        codic, dic = get_unit_dictionaries(self.unit)
        # print(unit, dic, codic)
        if unit in codic:
            scaling = codic[unit]
        elif unit in dic:
            scaling = dic[unit]
        else:
            raise RuntimeError('should not happen')

        if unit in codic:
            base_string = unit.to_string(format='unicode')
        elif unit in dic:
            base_string = unit.to_string(format='latex')[1:-1]
        scaling_str = str(Fraction(scaling).limit_denominator(10000))
        if scaling_str == '0':
            scaling_str = ''
        elif scaling_str == '1':
            scaling_str = base_string
        else:
            scaling_str = base_string + '^{' + scaling_str + '}'
        return scaling, scaling_str

    def _sanity_check(self, value):
        """
        Function for sanity-checking addition, subtraction, multiplication
        and division of quantities. They should all have same a and h.
        """
        # pylint: disable=protected-access
        err_msg = "Operation requires objects to have same a and h value."
        if isinstance(value, PaicosQuantity):
            if value._a != self._a:
                info = f' Obj1._a={self._a}, Obj2._a={value._a}'
                raise RuntimeError(err_msg + info)
            if value.h != self.h:
                info = f' Obj1.h={self.h}, Obj2.h={value.h}'
                raise RuntimeError(err_msg + info)

    def _repr_latex_(self):
        number_part = super()._repr_latex_().split('\\;')[0]
        _, pu_units = separate_units(self.unit)
        u_latex = (self.unit / pu_units).to_string(format='latex')[1:-1]
        pu_latex = pu_units.to_string(format='latex')[1:-1]

        if pu_units != u.Unit(''):
            modified = number_part + '\\;' + u_latex + \
                '\\times' + pu_latex + '$'
        else:
            modified = number_part + '\\;' + u_latex + '$'
        return modified

    def __add__(self, value):
        self._sanity_check(value)
        return super().__add__(value)

    def __sub__(self, value):
        self._sanity_check(value)
        return super().__sub__(value)

    def __mul__(self, value):
        self._sanity_check(value)
        return super().__mul__(value)

    def __truediv__(self, value):
        self._sanity_check(value)
        return super().__truediv__(value)

    def dump(self):
        """This astropy Quantity method has not been implemented the Paicos subclasses."""
        raise RuntimeError("not implemented")

    def dumps(self):
        """This astropy Quantity method has not been implemented the Paicos subclasses."""
        raise RuntimeError("not implemented")

    def tobytes(self):
        """This astropy Quantity method has not been implemented the Paicos subclasses."""
        raise RuntimeError("not implemented")

    def tofile(self):
        """This astropy Quantity method has not been implemented the Paicos subclasses."""
        raise RuntimeError("not implemented")

    def tolist(self):
        """This astropy Quantity method has not been implemented the Paicos subclasses."""
        raise RuntimeError("not implemented")

    def tostring(self):
        """This astropy Quantity method has not been implemented the Paicos subclasses."""
        raise RuntimeError("not implemented")

    def choose(self):
        """This astropy Quantity method has not been implemented the Paicos subclasses."""
        raise RuntimeError("not implemented")


class PaicosTimeSeries(PaicosQuantity):

    """
    PaicosTimeSeries is a subclass of the PaicosQuantity and and shares many
    of the same methods.

    The time series is very similar to the PaicosQuantity but
    the .a or .time properties now return arrays with same length
    as the object itself (.shape[0]).
    """

    def __new__(cls, value, unit=None, dtype=None, copy=True, order=None,
                subok=False, ndmin=0, h=None, a=None, comoving_sim=None):

        if isinstance(value, list):
            h = value[0].h  # Could check that they are all the same...
            comoving_sim = value[0].comoving_sim
            unit = value[0].unit
            dtype = value[0].dtype

            shapes = [value[i].value.shape for i in range(len(value))]
            err_msg = f'Not all shapes of the list are identical!, {shapes},\n\n'
            for shape in shapes:
                assert shape == shapes[0], err_msg + f'{shape}'

            a = np.array([value[i]._a for i in range(len(value))])
            value = np.array([value[i].value for i in range(len(value))])
        elif isinstance(value, np.ndarray):
            assert h is not None
            assert a is not None
            assert isinstance(a, np.ndarray)
            assert a.shape[0] == value.shape[0]
            if hasattr(value, 'unit'):
                if unit is not None:
                    raise RuntimeError(
                        'value has units but you are also passing unit={}:'.format(unit))
                unit = value.unit
                value = value.value
        else:
            raise RuntimeError('unexpected input for value:', value)

        msg = ('PaicosTimeSeries requires that the length of the first '
               + 'dimension is equal to the length of the time array')
        assert value.shape[0] == a.shape[0], msg

        assert len(value.shape) <= 3, 'Only 1D, 2D and 3D arrays are supported'

        obj = super().__new__(cls, value, unit=unit, dtype=dtype, copy=copy,
                              order=order, subok=subok, ndmin=ndmin, h=h, a=a,
                              comoving_sim=comoving_sim)

        obj._h = h
        obj._a = a
        obj._comoving_sim = comoving_sim

        return obj

    @property
    def to_physical(self):
        """
        Returns a copy of the current `PaicosTimeSeries` instance with the
        a and h factors removed, i.e. transform from comoving to physical.
        The value of the resulting object is scaled accordingly.
        """
        codic, _ = get_unit_dictionaries(self.unit)
        factor = self.h**codic[small_h]

        if self.comoving_sim:
            factor *= self.a**codic[small_a]

        value = self.view(np.ndarray)
        new_unit = get_new_unit(self.unit, [small_a, small_h])

        if len(value.shape) == 1:
            new_value = value * factor
        elif len(value.shape) == 2:
            new_value = value * factor[:, None]
        elif len(value.shape) == 3:
            new_value = value * factor[:, None, None]

        return self._new_view(new_value, new_unit)

    def to_comoving(self, unit):
        """
        Returns a copy of the current `PaicosTimeSeries` with the
        a and h factors given by the input unit.
        """

        if not self.comoving_sim:
            raise RuntimeError('Only implemented for comoving simulations')

        if isinstance(unit, str):
            unit = u.Unit(unit)

        u_unit_to, pu_unit_to = separate_units(unit)
        u_unit, pu_unit = separate_units(self.unit)

        if u_unit_to == u.Unit(''):
            u_unit_to = u_unit
        elif u_unit_to != u_unit:
            raise RuntimeError('This method can only change the a and h factors!')

        change_pu = pu_unit / pu_unit_to

        codic, _ = get_unit_dictionaries(change_pu)
        factor = self.h**codic[small_h]

        if self.comoving_sim:
            factor *= self.a**codic[small_a]

        value = self.view(np.ndarray)
        new_unit = self.unit / change_pu

        if len(value.shape) == 1:
            new_value = value * factor
        elif len(value.shape) == 2:
            new_value = value * factor[:, None]
        elif len(value.shape) == 3:
            new_value = value * factor[:, None, None]

        return self._new_view(new_value, new_unit)

    @property
    def hdf5_attrs(self):
        """
        Give the units as a dictionary for hdf5 data set attributes
        """
        return {'unit': self.unit.to_string(), 'Paicos': 'PaicosTimeSeries'}

    @property
    def copy(self):
        """
        Returns a copy of the PaicosQuantity
        """
        return PaicosTimeSeries(self.value, self.unit, a=self._a, h=self.h,
                                copy=True, comoving_sim=self.comoving_sim)

    def make_matrix(self, vec):
        """
        """
        assert vec.shape[0] == self.shape[0]
        return np.vstack([vec for _ in range(self.shape[1])]).T

    def _sanity_check(self, value):
        """
        Function for sanity-checking addition, subtraction, multiplication
        and division of quantities. They should all have same a and h.
        """
        err_msg = "Operation requires objects to have same a and h value.\n"
        if isinstance(value, PaicosQuantity):
            if not isinstance(value, PaicosTimeSeries):
                msg = ('operations combining PaicosQuantity and '
                       + 'PaicosTimeSeries is not allowed.')
                raise RuntimeError(msg)

        if isinstance(value, PaicosTimeSeries):
            info = f'\nObj1.a={self._a}.\n\nObj2.a={value._a}'
            try:
                np.testing.assert_array_equal(value._a, self._a)
            except AssertionError as exc:
                raise RuntimeError(err_msg + info) from exc
            if value.h != self.h:
                info = f'\nObj1.h={self.h}.\n\nObj2.h={value.h}'
                raise RuntimeError(err_msg + info)


def paicos_quantity_list_to_array(list_to_convert):
    """
    Only works on 1D lists...
    """
    e0 = list_to_convert[0]
    if hasattr(e0, 'unit'):
        for e in list_to_convert:
            assert e0.unit == e.unit
            assert e0.comoving_sim == e.comoving_sim
            assert e0._a == e._a
            assert e0._h == e._h
        return np.array([e.value for e in list_to_convert]) * e0.unit_quantity
    else:
        return np.array(list_to_convert)



"""
This module defines a class for 1D histograms.
"""
import time
import numpy as np
from .. import util
from .. import settings
from ..cython.histogram import get_hist_from_weights_and_idigit as func


def make_bins(bins, logscale):
    """
    Private method to calculate the edges and centers of the bins
    given lower, upper and number of bins.

    Parameters:
        bins (tuple): Tuple of lower edge, upper edge and number of
                      bins for the axis
    Returns:
        edges (array): Edges of the bins
        centers (array): Centers of the bins
    """

    lower, upper, nbins = bins
    edges, centers = make_bins_helper(lower, upper, nbins, logscale)
    if settings.use_units:
        assert lower.unit == upper.unit
        edges = np.array(edges) * lower.unit_quantity
        centers = np.array(centers) * lower.unit_quantity
    return edges, centers


@util.remove_astro_units
def make_bins_helper(lower, upper, nbins, logscale):
    """
    Private method to calculate the edges and centers of the bins
    in logscale or linear scale based on the class variable logscale
    Parameters:
        lower (float): Lower edge of the bin
        upper (float): Upper edge of the bin
        nbins (int): Number of bins for the axis
    Returns:
        edges (array): Edges of the bins
        centers (array): Centers of the bins
    """

    if logscale:
        lower = np.log10(lower)
        upper = np.log10(upper)

    edges = lower + np.arange(nbins + 1) * (upper - lower) / nbins
    centers = 0.5 * (edges[1:] + edges[:-1])

    if logscale:
        edges = 10**edges
        centers = 10**centers

    return edges, centers


class Histogram:
    """
    This is a brief class for efficient computation of many 1D histograms
    using the same x-variables and binning but different weights.

    The heavy part of the computation is stored in idigit
    """

    def __init__(self, x, bins, logscale=False, verbose=False):
        """
        Initialize the class.

        Parameters:
            x_variable (array): x-coordinates of the data.

            bins (array): Bin edges for the histogram.

            logscale (bool): Indicates whether to use logscale for the
                             histogram, default is True.

            verbose (bool, optional): If True, prints the time taken to
                                      digitize the x-coordinates.
        """

        if isinstance(bins, int):
            bins = [x.min(), x.max(), bins]

        self.logscale = logscale
        self.verbose = verbose

        self.edges, self.centers = make_bins(bins, self.logscale)
        self.bin_centers = self.centers
        self.bin_edges = self.edges

        if self.verbose:
            t = time.time()
            print('Digitize for histogram begun')

        self.idigit = np.digitize(x, self.edges)
        if self.verbose:
            print(f'This took {time.time() - t:1.2f} seconds')

    def hist(self, weights):
        """
        Compute the histogram of the data.

        Parameters:
            weights (array): Weights of the data.

        Returns:
            array: Histogram of the data.
        """

        get_hist_from_weights_and_idigit = util.remove_astro_units(func)

        # compute histogram using pre-digitized x-coordinates and given weights
        hist = get_hist_from_weights_and_idigit(self.edges.shape[0], weights,
                                                self.idigit)
        if settings.use_units:
            hist = hist * weights.unit_quantity

        return hist



"""This defines a reader for Arepo snapshot files"""
import time
from inspect import signature
import numbers
import warnings
import numpy as np
import h5py
from .arepo_catalog import Catalog
from .paicos_readers import PaicosReader
from ..writers.paicos_writer import PaicosWriter
from .. import settings
from ..derived_variables import derived_variables


class Snapshot(PaicosReader):
    """
    This is a Python class for reading Arepo snapshots, which are simulations
    of the evolution of the universe using a code called Arepo. The class is
    based on a script originally written by Ewald Puchwein, which has since
    then been modified and included in Paicos.

    The class takes in the path of the directory containing the snapshot, the
    snapshot number, and an optional basename parameter, and uses this
    information to locate and open the snapshot files. The class also loads
    the snapshot's header, parameters, and configuration, and uses them to
    create a converter object that can be used to convert units in the
    snapshot. The class also includes methods to extract the redshift, scale
    factor, and other properties of the snapshot, as well as the subfind
    catalog if present.

    Important methods and attributes.

    snap = Snapshot()

    snap.Parameters (dict): Contains information from the parameter
                            file used in the simulation (e.g. param.txt).

    snap.Config (dict): Contains information from the Config
                        file used in the simulation (e.g. Config.txt).

    snap.Header (dict): Contains information about this particular snapshot
                        such as its time (e.g scale factor).

    snap.z (float): redshift

    snap.h (float): reduced Hubble param (e.g. 0.67)

    snap.age: the age of the Universe (only for cosmological runs)
    snap.lookback_time: the age of the Universe (only for cosmological runs)

    snap.time: the time stamp of the snapshot (only for non-cosmological runs)

    snap.box_size: the dimensions of the simulation domain

    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, basedir, snapnum=None, basename="snap", load_all=False,
                 to_physical=False, load_catalog=None, verbose=False):
        """
        Initialize the Snapshot class.

        Parameters:

        basedir (str): path of the directory containing the snapshot
                       (e.g. the 'output' folder)

        snapnum (int): snapshot number

        basename (str): name of the snapshot file, default is "snap"

        verbose (bool): whether to print information about the snapshot,
        default is False

        no_snapdir (bool): whether there is no snap directory, i.e.,
                           default is False

        load_catalog (bool): whether to load the subfind catalog.
                             The default None is internally changed to
                             True for comoving simulations and to
                             False for non-comoving simulations.
    """

        super().__init__(basedir=basedir, snapnum=snapnum, basename=basename,
                         load_all=load_all, to_physical=to_physical,
                         basesubdir='snapdir', verbose=verbose)

        assert self.snapnum is not None

        self.load_catalog = load_catalog

        if not hasattr(self, "dic_selection_index"):
            self.dic_selection_index = {}

        self.nfiles = self.Header["NumFilesPerSnapshot"]
        self.npart = self.Header["NumPart_Total"]
        self.nspecies = self.npart.size

        if self.verbose:
            print("has", self.nspecies, "particle types")
            print("with npart =", self.npart)

        self.box = self.Header["BoxSize"]
        box_size = [self.box, self.box, self.box]

        for ii, dim in enumerate(['X', 'Y', 'Z']):
            if 'LONG_' + dim in self.Config:
                box_size[ii] *= self.Config['LONG_' + dim]

        if settings.use_units:
            get_paicos_quantity = self.get_paicos_quantity
            self.box_size = get_paicos_quantity(box_size, 'Coordinates')
            self.masstable = get_paicos_quantity(self.Header["MassTable"],
                                                 'Masses')
        else:
            self.box_size = np.array(box_size)
            self.masstable = self.Header["MassTable"]

        # get subfind catalog?
        if load_catalog is None:
            load_catalog = bool(self.ComovingIntegrationOn)

        if load_catalog:
            try:
                self.Cat = Catalog(
                    self.basedir, self.snapnum, verbose=self.verbose,
                    subfind_catalog=True)
            except FileNotFoundError:
                self.Cat = None

            # If no subfind catalog found, then try for a fof catalog
            if self.Cat is None:
                try:
                    self.Cat = Catalog(
                        self.basedir, self.snapnum, verbose=self.verbose,
                        subfind_catalog=False)
                except FileNotFoundError:
                    warnings.warn('no catalog found')

        self.P_attrs = {}  # attributes

        self.derived_data_counter = 0

        self._add_mass_to_user_funcs()

        self._find_available_for_loading()
        self._find_available_functions()

        self._identify_parttypes()

        self.__get_auto_comple_list()

    def _add_mass_to_user_funcs(self):
        """
        This functions adds functionality for obtaining the masses
        of particle types which do not have the Masses blockname stored.
        These normally instead have their mass in masstable.
        """

        self._this_snap_funcs = {}

        class Mass:
            """
            This class allows us to get a function which is just a function of
            one parameter.
            """
            def __init__(self, parttype):
                """
                The parttype, e.g. 1 for DM particles
                """
                self.parttype = parttype

            def get_masses_from_header(self, snap):
                """
                Get mass of particle type from the mass table
                """
                parttype = self.parttype
                if parttype in snap.dic_selection_index:
                    npart = snap.dic_selection_index[parttype].shape[0]
                else:
                    npart = snap.npart[parttype]
                return np.ones(npart) * snap.masstable[parttype]

        # Add function to the ones available
        for parttype in range(self.nspecies):
            if self.masstable[parttype] != 0:
                p_key = str(parttype) + '_Masses'
                obj = Mass(parttype)
                self._this_snap_funcs[p_key] = obj.get_masses_from_header

    def _find_available_for_loading(self):
        """
        Read the hdf5 file info and find all the blocknames
        that are available for each particle type.
        """
        self._all_avail_load = []
        self._part_avail_load = {i: [] for i in range(self.nspecies)}
        self._part_specs = {i: {} for i in range(self.nspecies)}
        for parttype in range(self.nspecies):
            parttype_str = f'PartType{parttype}'

            if not self.multi_file:
                with h5py.File(self.filename, 'r') as file:
                    if parttype_str in file:
                        for key in file[parttype_str]:
                            p_key = f'{parttype}_{key}'
                            self._part_avail_load[parttype].append(p_key)
                            shape = file[parttype_str][key].shape
                            dtype = file[parttype_str][key].dtype
                            self._part_specs[parttype][key] = {'shape': shape,
                                                               'dtype': dtype}
            else:
                # For multiple files we sometimes need to open many of them
                # to find *all* the available parttypes.
                # This can occur when the parttypes are not evenly distributed
                # in space (e.g. the high resolution region in a zoom
                # will have no low res DM particles).
                for ii in range(self.nfiles):
                    with h5py.File(self.multi_filename.format(ii), 'r') as file:
                        if parttype_str in file:
                            for key in file[parttype_str]:
                                p_key = f'{parttype}_{key}'
                                self._part_avail_load[parttype].append(p_key)
                                shape = list(file[parttype_str][key].shape)
                                if len(shape) > 1:
                                    shape[0] = self.npart[parttype]
                                dtype = file[parttype_str][key].dtype
                                self._part_specs[parttype][key] = {'shape': tuple(shape),
                                                                   'dtype': dtype}
                            # print(f'found parttype in partfile {ii}, breaking out')
                            break

            self._all_avail_load += self._part_avail_load[parttype]

    def _identify_parttypes(self):
        """
        Try to figure out which physical variable is stored in each
        particle type.
        """
        self._type_info = {0: 'voronoi_cells'}
        for p in range(1, self.nspecies):
            bh = any('BH_' in key for key in self._part_avail_load[p])
            star = any('GFM_' in key for key in self._part_avail_load[p])
            if bh:
                self._type_info[p] = 'black_holes'
            if star:
                self._type_info[p] = 'stars'

    def _find_available_functions(self):
        """
        This function goes through all the implemented functions
        for getting derived variables. Checking their dependencies,
        it then figures out which derived variables are actually possible
        for this particular snapshot. For instance, you can calculate the
        magnetic field strength if the magnetic field is not stored in
        the hdf5 file.
        """

        user_functs = derived_variables.user_functions

        for func_name, func in user_functs.items():
            self._this_snap_funcs[func_name] = func

        # Add all implemented functions
        if not settings.use_only_user_functions:
            def_functs = derived_variables.default_functions
            for func_name, func in def_functs.items():
                if func_name not in self._this_snap_funcs:
                    self._this_snap_funcs[func_name] = func

        # Build a dependency dictionary by asking each function for its
        # dependencies
        self._dependency_dic = {}
        for func_name, func in self._this_snap_funcs.items():
            sig = signature(func)
            # If the function has two input arguments, then we assume
            # that passing True to the second argument will return its
            # dependencies
            if len(sig.parameters) == 2:
                self._dependency_dic[func_name] = func(self, True)
            else:
                self._dependency_dic[func_name] = []

        dependency_dic = dict(self._dependency_dic)

        # We then trim the dependency dictionary

        # This will fail for very nested dependencies.
        for _ in range(3):
            # First substitute all dependencies that are
            # at the top level of the dictionary
            for func_name, deps in dependency_dic.items():
                for dep in list(deps):
                    if dep in dependency_dic:
                        deps.remove(dep)
                        for subdep in dependency_dic[dep]:
                            deps.append(subdep)

            # Then remove all the dependencies that can be loaded
            for func_name, deps in dependency_dic.items():
                for dep in list(deps):
                    if dep in self._all_avail_load:
                        deps.remove(dep)

        # Delete the entries where we do not have the requirements
        for func_name, deps in dependency_dic.items():
            if len(deps) > 0:
                if func_name in user_functs:
                    msg = (f'Deleting the user function: {user_functs[func_name]} '
                           + f'because its dependency: {deps} '
                           + 'is missing')
                    warnings.warn(msg)
                del self._this_snap_funcs[func_name]

    def get_variable_function(self, p_key, info=False):
        """
        This is a helper function for 'get_derived_data'. It returns a
        function if info is False, and a list of all available
        functions for parttype = p_key[0] if info is True.
        """

        assert isinstance(p_key, str)

        if not p_key[0].isnumeric() or p_key[1] != '_':
            msg = (f"The key '{p_key}' is not valid."
                   + "\n\nKeys are expected to consist of an integer "
                   + "(the particle type) and a blockname, separated by a "
                   + " _. For instance 0_Density. You can get the "
                   + "available fields like so: snap.info(0)")
            raise RuntimeError(msg)

        # Return a list with all available keys for this parttype
        if info:
            parttype = int(p_key[0])
            avail_list = []
            for key in self._this_snap_funcs:
                if int(key[0]) == parttype:
                    avail_list.append(key)

            return avail_list

        # Return a function
        if p_key in self._this_snap_funcs:
            return self._this_snap_funcs[p_key]
        msg = '\n\n{} not found in the functions: {}'
        msg = msg.format(p_key, self._this_snap_funcs)
        raise RuntimeError(msg)

    def info(self, parttype, verbose=True):
        """
        This function provides information about the keys of a certain
        particle type in a snapshot.

        Args: PartType (int): An integer representing the particle type of
        interest. verbose (bool, optional): A flag indicating whether or not
        to print the keys to the console. Defaults to True.

        Returns: list or None : If the PartType exists in the file, a list of
        keys for that PartType is returned, otherwise None.
        """
        parttype_str = f'PartType{parttype}'
        if parttype >= self.nspecies:
            err_msg = (f"Parttype {self.nspecies-1} is the largest contained"
                       + f" in the snapshot which has nspecies={self.nspecies}.")
            raise RuntimeError(err_msg)

        load_keys = self._part_avail_load[parttype]
        if verbose:
            print('\nKeys for ' + parttype_str + ' in the hdf5 file:')
            for key in (sorted(load_keys)):
                print_key = key
                if settings.use_aliases:
                    if key in settings.aliases:
                        alias = settings.aliases[key]
                        msg = alias + '\t' * 5 + '(an alias of {})'
                        print_key = msg.format(key)

                print(print_key)

            print('\nPossible derived variables are:')
            dkeys = self.get_variable_function(f'{parttype}_', True)

            for key in (sorted(dkeys)):
                print_key = key
                if settings.use_aliases:
                    if key in settings.aliases:
                        alias = settings.aliases[key]
                        msg = alias + '\t' * 5 + '(an alias of {})'
                        print_key = msg.format(key)

                print(print_key)
        else:
            return load_keys

    def load_data_experimental(self, parttype, blockname):
        """
        Load data from hdf5 file(s). Example usage:

        snap = Snapshot(...)

        snap.load_data(0, 'Density')

        Note that subsequent calls does not reload the data. Reloading
        the data can be done explicitly:

        snap.remove_data(0, 'Density')
        snap.load_data(0, 'Density')

        """

        assert parttype < self.nspecies

        p_key = f'{parttype}_{blockname}'
        alias_key = p_key

        if settings.use_aliases:
            if p_key in settings.aliases:
                alias_key = settings.aliases[p_key]

        if p_key not in self.info(parttype, False):
            msg = (f'Unable to load parttype {parttype}, blockname '
                   + f'{blockname} as this field is not in the hdf5 file')
            raise RuntimeError(msg)

        if alias_key in self:
            if self.verbose:
                print(blockname, "for species",
                      parttype, "already in memory")
            return

        if self.verbose:
            print("loading block", blockname,
                  "for species", parttype, "...")
            start_time = time.time()

        def read_hdf5_file(filename, parttype, blockname):
            """
            This helper function does the actual data loading
            """
            parttype_str = f'PartType{parttype}'
            with h5py.File(filename, 'r') as f:
                # Load a dataset
                dataset = f[parttype_str][blockname][()]

            if settings.double_precision:
                # Load all variables with double precision
                if not issubclass(dataset.dtype.type, numbers.Integral):
                    dataset = dataset.astype(np.float64)
            else:
                warnings.warn('\n\nThe cython routines expect double precision '
                              + 'and will fail unless settings.double_precision '
                              + 'is True.\n\n')

            if settings.use_units:
                if parttype in self._type_info:
                    ptype = self._type_info[parttype]  # e.g. 'voronoi_cells'
                    dataset = self.get_paicos_quantity(dataset, blockname,
                                                       field=ptype)
                else:
                    # Assume dark matter for the units
                    dataset = self.get_paicos_quantity(dataset, blockname,
                                                       field='dark_matter')
            return dataset

        if self.multi_file:
            filenames = [self.multi_filename.format(ii) for ii in range(self.nfiles)]
            shape = self._part_specs[parttype][blockname]['shape']
            if len(shape) == 1:
                self[alias_key] = np.hstack([read_hdf5_file(filename, parttype, blockname)
                                             for filename in filenames])
            elif len(shape) == 2:
                self[alias_key] = np.vstack([read_hdf5_file(filename, parttype, blockname)
                                             for filename in filenames])
            else:
                raise RuntimeError(f"Unexpected shape={shape} for {parttype}_{blockname}")
        else:
            self[alias_key] = read_hdf5_file(self.filename, parttype, blockname)

        # Only keep the cells with True in the selection index array
        if parttype in self.dic_selection_index:
            selection_index = self.dic_selection_index[parttype]
            shape = self[alias_key].shape
            if len(shape) == 1:
                self[alias_key] = self[alias_key][selection_index]
            elif len(shape) == 2:
                self[alias_key] = self[alias_key][selection_index, :]
            else:
                raise RuntimeError('Data has unexpected shape!')

        if self.verbose:
            print("... done! (took", time.time() - start_time, "s)")

    def load_data(self, parttype, blockname):
        """
        Load data from hdf5 file(s). Example usage:

        snap = Snapshot(...)

        snap.load_data(0, 'Density')

        Note that subsequent calls does not reload the data. Reloading
        the data can be done explicitly:

        snap.remove_data(0, 'Density')
        snap.load_data(0, 'Density')

        """

        assert parttype < self.nspecies

        p_key = f'{parttype}_{blockname}'
        alias_key = p_key

        if settings.use_aliases:
            if p_key in settings.aliases:
                alias_key = settings.aliases[p_key]

        if p_key not in self.info(parttype, False):
            msg = (f'Unable to load parttype {parttype}, blockname '
                   + f'{blockname} as this field is not in the hdf5 file')
            raise RuntimeError(msg)

        datname = f'PartType{parttype}/{blockname}'
        if alias_key in self:
            if self.verbose:
                print(blockname, "for species",
                      parttype, "already in memory")
            return

        if self.verbose:
            print("loading block", blockname,
                  "for species", parttype, "...")
            start_time = time.time()

        skip_part = 0

        for ifile in range(self.nfiles):
            if self.multi_file is False:
                cur_filename = self.filename
            else:
                if self.no_subdir:
                    cur_filename = self.multi_wo_dir.format(ifile)
                else:
                    cur_filename = self.multi_filename.format(ifile)

            f = h5py.File(cur_filename, "r")

            np_file = f["Header"].attrs["NumPart_ThisFile"][parttype]

            if ifile == 0:   # initialize array
                shape = self._part_specs[parttype][blockname]['shape']
                dtype = self._part_specs[parttype][blockname]['dtype']
                if len(shape) == 1:
                    self[alias_key] = np.empty(self.npart[parttype], dtype=dtype)
                else:
                    self[alias_key] = np.empty((self.npart[parttype], shape[1]),
                                               dtype=dtype)
            if np_file > 0:
                self[alias_key][skip_part:skip_part + np_file] = f[datname]

            skip_part += np_file

        if settings.double_precision:
            # Load all variables with double precision
            if not issubclass(self[alias_key].dtype.type, numbers.Integral):
                self[alias_key] = self[alias_key].astype(np.float64)
        else:
            warnings.warn('\n\nThe cython routines expect double precision '
                          + 'and will fail unless settings.double_precision '
                          + 'is True.\n\n')

        # Only keep the cells with True in the selection index array
        if parttype in self.dic_selection_index:
            selection_index = self.dic_selection_index[parttype]
            shape = self[alias_key].shape
            if len(shape) == 1:
                self[alias_key] = self[alias_key][selection_index]
            elif len(shape) == 2:
                self[alias_key] = self[alias_key][selection_index, :]
            else:
                raise RuntimeError('Data has unexpected shape!')

        if settings.use_units:
            if parttype in self._type_info:
                ptype = self._type_info[parttype]  # e.g. 'voronoi_cells'
                self[alias_key] = self.get_paicos_quantity(self[alias_key],
                                                           blockname,
                                                           field=ptype)
            else:
                # Assume dark matter for the units
                self[alias_key] = self.get_paicos_quantity(self[alias_key],
                                                           blockname,
                                                           field='dark_matter')
            if not hasattr(self[alias_key], 'unit'):
                del self[alias_key]

        if self.verbose:
            print("... done! (took", time.time() - start_time, "s)")

    def get_derived_data(self, parttype, blockname, verbose=False):
        """
        Get derived quantities. Example usage:

        snap = Snapshot(...)

        snap.get_derived_data(0, 'Temperatures')

        """
        # from .derived_variables import get_variable_function

        p_key = str(parttype) + "_" + blockname

        msg = (f'\n\n{blockname} is in the hdf5 file(s), please use load_data '
               + 'instead of get_derived_data')
        assert p_key not in self.info(parttype, False), msg

        if verbose:
            msg1 = f'Attempting to get derived variable: {p_key}...'
            msg2 = f'So we need the variable: {p_key}...'
            if self.derived_data_counter == 0:
                print(msg1, end='')
            else:
                print('\n\t' + msg2, end='')
            self.derived_data_counter += 1

        func = self.get_variable_function(p_key)

        if settings.use_aliases:
            if p_key in settings.aliases:
                p_key = settings.aliases[p_key]
        self[p_key] = func(self)

        if verbose:
            self.derived_data_counter -= 1
            if self.derived_data_counter == 0:
                print('\t[DONE]\n')

    def __getitem__(self, p_key):
        """
        This method is a special method in Python classes, known as a "magic
        method" that allows instances of the class to be accessed like a
        dictionary, using the bracket notation.

        This method is used to access the data stored in the class, it takes a
        single argument:

        p_key : a string that represents the data that is being accessed, it
        should be in the format of parttype_name, where parttype is an integer
        and name is the name of the data block. It first checks if the key is
        already in the class, if not it checks if the key is in the format of
        parttype_name and if the parttype exists in the file and the name is a
        valid key in the hdf5 file. If it is, the method loads the data from
        the hdf5 file, otherwise it calls the get_derived_data method to
        calculate the derived data. If the key does not meet the expected
        format or the parttype or data block does not exist it raises a
        RuntimeError with a message explaining the expected format.

        The method returns the value of the data block if it is found or
        loaded, otherwise it raises an error.

        This method allows for easy and convenient access to the data stored
        in the class, without the need for explicit method calls to load or
        calculate the data.

        That is, one can do snap['0_Density'] and the data will automatically
        be loaded.
        """

        if settings.use_aliases:

            if p_key in settings.inverse_aliases:
                p_key = settings.inverse_aliases[p_key]

        if p_key not in self:
            if not p_key[0].isnumeric() or p_key[1] != '_':
                msg = (f"The key '{p_key}' is not valid."
                       + "\n\nKeys are expected to consist of an integer "
                       + "(the particle type) and a blockname, separated by a "
                       + " _. For instance 0_Density. You can get the "
                       + "available fields like so: snap.info(0)")
                raise RuntimeError(msg)
            parttype = int(p_key[0])
            name = p_key[2:]

            if parttype >= self.nspecies:
                msg = f'Simulation only has {self.nspecies} species.'
                raise RuntimeError(msg)

            if p_key in self.info(parttype, False):
                self.load_data(parttype, name)
            else:
                verbose = settings.print_info_when_deriving_variables
                self.get_derived_data(parttype, name, verbose=verbose)

        if settings.use_aliases:
            if p_key in settings.aliases:
                p_key = settings.aliases[p_key]
        return super().__getitem__(p_key)

    def __get_auto_comple_list(self):
        """
        Pre-compute a list for auto-completion.
        """
        self._auto_list = list(self._all_avail_load)
        for key in self._this_snap_funcs:
            self._auto_list.append(key)

        if settings.use_aliases:
            for ii, p_key in enumerate(self._auto_list):
                if p_key in settings.aliases:
                    self._auto_list[ii] = settings.aliases[p_key]

    def _ipython_key_completions_(self):
        """
        Auto-completion of dictionary.
        """

        return self._auto_list

    def remove_data(self, parttype, blockname):
        """
        Remove data from object. Sometimes useful for for large datasets
        """
        p_key = str(parttype) + "_" + blockname
        if p_key in self:
            del self[p_key]
        if p_key in self.P_attrs:
            del self.P_attrs[p_key]

    def select(self, selection_index, parttype=None):
        """
        Create a new snapshot object which will only contain
        cells with a selection_index.

        Example use:
        index = snap['0_Density'] > snap['0_Density'].unit_quantity*1e-6
        selected_snap = snap.select(index, parttype=0)

        """
        if parttype is None:
            parttype = 0
            msg = ("You have not specified the parttype that the selection index"
                   + " corresponds to. The code will assume parttype=0. Future versions"
                   + " of this method will likely have parttype as a required input")
            warnings.warn(msg)

        s_index = selection_index

        if parttype in self.dic_selection_index:
            # This snap object is already a selection, combine the criteria!
            previous_selection = self.dic_selection_index[parttype]
            new_index = previous_selection[s_index]
            dic_selection_index = dict(self.dic_selection_index)
            dic_selection_index[parttype] = new_index
        else:
            # Convert to integer array
            if s_index.dtype == 'bool':
                s_index = np.arange(s_index.shape[0])[s_index]

            dic_selection_index = dict(self.dic_selection_index)
            dic_selection_index[parttype] = s_index

        select_snap = Snapshot(self.basedir, self.snapnum,
                               basename=self.basename,
                               verbose=self.verbose,
                               to_physical=self.to_physical,
                               load_catalog=self.load_catalog)

        select_snap.dic_selection_index = dic_selection_index

        for key in self:
            if key[0] == str(parttype):
                shape = self[key].shape
                if shape[0] == 1 and s_index.shape[0] != 1:
                    # Copy over single float numbers, e.g., constants,
                    # such as the mean molecular weight
                    select_snap[key] = self[key]
                else:
                    if len(shape) == 1:
                        select_snap[key] = self[key][s_index]
                    elif len(shape) == 2:
                        select_snap[key] = self[key][s_index, :]
                    else:
                        raise RuntimeError('Data has unexpected shape!')
            else:
                select_snap[key] = self[key]

        return select_snap

    def radial_select(self, center, r_max, r_min=0.0, parttype=None):
        """
        A convenient function for selecting in radius

        Returns a new snapshot with the radial selection
        """
        from .. import util
        if parttype is None:

            index = util.get_index_of_radial_range(self['0_Coordinates'],
                                                   center, r_min, r_max)

            selected_snap = self.select(index, parttype=0)
            for p in range(1, self.nspecies):
                index = util.get_index_of_radial_range(selected_snap[f'{p}_Coordinates'],
                                                       center, r_min, r_max)
                selected_snap = selected_snap.select(index, parttype=p)

        else:
            p = parttype
            index = util.get_index_of_radial_range(self[f'{p}_Coordinates'],
                                                   center, r_min, r_max)
            selected_snap = self.select(index, parttype=p)

        return selected_snap

    def save_new_snapshot(self, basename, single_precision=False):
        """
        Save a new snapshot containing the currently loaded (derived)
        variables. Useful for reducing datasets to smaller sizes.
        """
        writer = PaicosWriter(self, self.basedir, basename, 'w')

        new_npart = [0] * self.nspecies
        for key in self:
            for parttype in range(self.nspecies):
                if key[:2] == f'{parttype}_':
                    new_npart[parttype] = self[key].shape[0]
                    parttype_str = f'PartType{parttype}'

                    if single_precision:
                        data = self[key].astype(np.float32)
                    else:
                        data = self[key]
                    writer.write_data(key[2:], data, group=parttype_str)

        with h5py.File(writer.tmp_filename, 'r+') as f:
            f['Header'].attrs["NumFilesPerSnapshot"] = 1
            f['Header'].attrs["NumPart_Total"] = np.array(new_npart)

        writer.finalize()

        return writer

    def get_sum_of_array(self, variable):
        """
        Find the sum of a 1D array.

        Equivalent to np.sum(arr, axis=0) but with open_mp
        """
        from .. import util

        if isinstance(variable, str):
            variable = self[variable]
        else:
            if not isinstance(variable, np.ndarray):
                raise RuntimeError('Unexpected type for variable')

        if len(variable.shape) == 1:
            from ..cython.get_derived_variables import sum_1d_array_omp as sum_omp
        elif len(variable.shape) == 2:
            from ..cython.get_derived_variables import sum_2d_array_omp as sum_omp
        else:
            raise RuntimeError('unexpected')

        if settings.use_units:
            sum_omp = util.remove_astro_units(sum_omp)
            uq = variable.uq

            return sum_omp(variable, settings.numthreads_reduction) * uq
        return sum_omp(variable, settings.numthreads_reduction)

    def get_sum_of_array_times_vector(self, arr, vector):
        """
        Find the sum of a 1D array times a 2D array

        Equivalent to np.sum(arr[:, None] * vector, axis=0) but with open_mp
        """
        from .. import util
        from ..cython.get_derived_variables import sum_arr_times_vector_omp
        numthreads = settings.numthreads_reduction

        if isinstance(arr, str):
            arr = self[arr]
        else:
            if not isinstance(arr, np.ndarray):
                raise RuntimeError('Unexpected type for variable')

        if isinstance(vector, str):
            vector = self[vector]
        else:
            if not isinstance(vector, np.ndarray):
                raise RuntimeError('Unexpected type for variable')

        assert vector.shape[1] == 3, 'only works on vectors, e.g. coordinates'

        if settings.use_units:
            sum_arr_times_vector_omp = util.remove_astro_units(sum_arr_times_vector_omp)
            uq = arr.uq * vector.uq

            return sum_arr_times_vector_omp(arr, vector, numthreads) * uq
        return sum_arr_times_vector_omp(arr, vector, numthreads)

    def get_sum_of_arr_times_vector_cross_product(self, mass, coord, velocity, center):
        """
        This code calculates sum_i (mass_i (coord_ij - center) x velocity_ij).

        That is, it returns the total angular momentum vector.
        """
        from .. import util
        from ..cython.get_derived_variables import sum_arr_times_vector_cross_product
        numthreads = settings.numthreads_reduction

        if isinstance(mass, str):
            mass = self[mass]
        else:
            if not isinstance(mass, np.ndarray):
                raise RuntimeError('Unexpected type for variable')

        if isinstance(coord, str):
            coord = self[coord]
        else:
            if not isinstance(coord, np.ndarray):
                raise RuntimeError('Unexpected type for variable')

        if isinstance(velocity, str):
            velocity = self[velocity]
        else:
            if not isinstance(velocity, np.ndarray):
                raise RuntimeError('Unexpected type for variable')

        if settings.use_units:
            sum_func = util.remove_astro_units(sum_arr_times_vector_cross_product)
            uq = mass.uq * coord.uq * velocity.uq

            return sum_func(mass, coord, velocity, center, numthreads) * uq
        return sum_func(mass, coord, velocity, center, numthreads)

    def center_of_mass(self, parttype=None):
        """
        Finds the center of mass for the entire snapshot.

        parttype (default None): if None, then all parttypes are included in the
                                 center-of-mass calculation. If e.g. parttype=0,
                                 then the center of mass of the gas is returned.
                                 if parttype=None, then the total center and
                                 a list of of the parttype centers are returned

        Can be used in combination with self.select to find the
        center of mass of a selection.
        """
        if parttype is None:
            mass = [self.get_sum_of_array(f'{p}_Masses') for p in range(self.nspecies)]
            mcord = [self.get_sum_of_array_times_vector(f'{p}_Masses', f'{p}_Coordinates')
                     for p in range(self.nspecies)]
            centers = [mcord[p] / mass[p] for p in range(self.nspecies)]

            tot_center = np.sum(np.vstack(mcord), axis=0) / np.sum(np.stack(mass))

            return tot_center, centers
        else:
            p = parttype
            mcord = self.get_sum_of_array_times_vector(f'{p}_Masses', f'{p}_Coordinates')
            mass = self.get_sum_of_array(f'{p}_Masses')
            return mcord / mass

    def total_angular_momentum(self, center, parttype=None):
        """
        Finds the total angular momentum for the entire snapshot.

        center: the center around which to calculate the angular momentum

        parttype (default None): if None, then all parttypes are included in the
                                 calculation. If e.g. parttype=0,
                                 then total angular momentum of the gas is returned.

        Can be used in combination with self.select to find the
        total angular momentum of a selection.
        """
        func = self.get_sum_of_arr_times_vector_cross_product
        if parttype is None:
            if settings.use_units:
                for p in range(self.nspecies):
                    assert self[f'{p}_Coordinates'].unit == center.unit
            res = [func(f'{p}_Masses', f'{p}_Coordinates', f'{p}_Velocities', center)
                   for p in range(self.nspecies)]
            return np.sum(np.vstack(res), axis=0), res
        else:
            p = parttype
            if settings.use_units:
                assert self[f'{p}_Coordinates'].unit == center.unit
            return func(f'{p}_Masses', f'{p}_Coordinates', f'{p}_Velocities', center)



from cython.parallel import prange
cimport openmp
import numpy as np
cimport numpy as np

ctypedef fused real_t:
    float
    double
    np.float32_t
    np.float64_t


def get_cube(real_t [:, :] pos, real_t xc, real_t yc, real_t zc,
             real_t sidelength_x, real_t sidelength_y,
             real_t thickness, real_t boxsize, int numthreads):

    cdef int Np = pos.shape[0]
    cdef int ip
    cdef real_t x, y, z

    cdef int[:] index = np.zeros(Np, dtype=np.intc)

    openmp.omp_set_num_threads(numthreads)

    for ip in prange(Np, nogil=True, schedule='static'):
        x = pos[ip, 0] - xc
        y = pos[ip, 1] - yc
        z = pos[ip, 2] - zc

        # # box_wrap_diff function
        # if x < -0.5*boxsize:
        #     x = x + boxsize
        # elif x > 0.5*boxsize:
        #     x = x - boxsize

        # if y < -0.5*boxsize:
        #     y = y + boxsize
        # elif y > 0.5*boxsize:
        #     y = y - boxsize

        # if z < -0.5*boxsize:
        #     z = z + boxsize
        # elif z > 0.5*boxsize:
        #     z = z - boxsize

        # Index calculation
        index[ip] = 0
        if (x < sidelength_x/2.0) and (x > -sidelength_x/2):
            if (y < sidelength_y/2.0) and (y > -sidelength_y/2):
                if (z > -0.5*thickness):
                    if (z < 0.5*thickness):
                        index[ip] = 1

    # Return a numpy boolean array
    tmp = np.zeros(Np, dtype=np.bool_)
    tmp[:] = index[:]
    return tmp


def get_cube_plus_thin_layer(real_t [:, :] pos, real_t xc, real_t yc, real_t zc,
                             real_t sidelength_x, real_t sidelength_y,
                             real_t sidelength_z, real_t [:] thickness,
                             real_t boxsize, int numthreads):

    cdef int Np = pos.shape[0]
    cdef int ip
    cdef real_t x, y, z

    cdef int[:] index = np.zeros(Np, dtype=np.intc)

    openmp.omp_set_num_threads(numthreads)

    for ip in prange(Np, nogil=True, schedule='static'):
        x = pos[ip, 0] - xc
        y = pos[ip, 1] - yc
        z = pos[ip, 2] - zc

        # # box_wrap_diff function
        # if x < -0.5*boxsize:
        #     x = x + boxsize
        # elif x > 0.5*boxsize:
        #     x = x - boxsize

        # if y < -0.5*boxsize:
        #     y = y + boxsize
        # elif y > 0.5*boxsize:
        #     y = y - boxsize

        # if z < -0.5*boxsize:
        #     z = z + boxsize
        # elif z > 0.5*boxsize:
        #     z = z - boxsize

        # Index calculation
        index[ip] = 0
        if (x < (sidelength_x + thickness[ip])/2.0) and (x > -(sidelength_x + thickness[ip])/2.0):
            if (y < (sidelength_y + thickness[ip])/2.0) and (y > -(sidelength_y + thickness[ip])/2.0):
                if (z < (sidelength_z + thickness[ip])/2.0):
                    if (z > -(sidelength_z + thickness[ip])/2.0):
                        index[ip] = 1

    # Return a numpy boolean array
    tmp = np.zeros(Np, dtype=np.bool_)
    tmp[:] = index[:]
    return tmp

def get_rotated_cube(real_t [:, :] pos, real_t xc, real_t yc, real_t zc,
                             real_t sidelength_x, real_t sidelength_y,
                             real_t sidelength_z,
                             real_t boxsize,
                             real_t[:] unit_vector_x,
                             real_t[:] unit_vector_y,
                             real_t[:] unit_vector_z,
                             int numthreads):

    cdef int Np = pos.shape[0]
    cdef int ip
    cdef real_t x, y, z
    cdef real_t x_dot_ex, x_dot_ey, x_dot_ez

    cdef int[:] index = np.zeros(Np, dtype=np.intc)

    openmp.omp_set_num_threads(numthreads)

    for ip in prange(Np, nogil=True, schedule='static'):
        x = pos[ip, 0] - xc
        y = pos[ip, 1] - yc
        z = pos[ip, 2] - zc

        # # box_wrap_diff function
        # if x < -0.5*boxsize:
        #     x = x + boxsize
        # elif x > 0.5*boxsize:
        #     x = x - boxsize

        # if y < -0.5*boxsize:
        #     y = y + boxsize
        # elif y > 0.5*boxsize:
        #     y = y - boxsize

        # if z < -0.5*boxsize:
        #     z = z + boxsize
        # elif z > 0.5*boxsize:
        #     z = z - boxsize

        x_dot_ex = x * unit_vector_x[0] + y * unit_vector_x[1] + z * unit_vector_x[2]
        x_dot_ey = x * unit_vector_y[0] + y * unit_vector_y[1] + z * unit_vector_y[2]
        x_dot_ez = x * unit_vector_z[0] + y * unit_vector_z[1] + z * unit_vector_z[2]

        # Index calculation
        index[ip] = 0
        if (x_dot_ex < sidelength_x/2.0) and (x_dot_ex > -sidelength_x/2.0):
            if (x_dot_ey < sidelength_y/2.0) and (x_dot_ey > -sidelength_y/2.0):
                if (x_dot_ez < sidelength_z/2.0):
                    if (x_dot_ez > -sidelength_z/2.0):
                        index[ip] = 1

    # Return a numpy boolean array
    tmp = np.zeros(Np, dtype=np.bool_)
    tmp[:] = index[:]
    return tmp

def get_rotated_cube_plus_thin_layer(real_t [:, :] pos, real_t xc, real_t yc, real_t zc,
                             real_t sidelength_x, real_t sidelength_y,
                             real_t sidelength_z, real_t [:] thickness,
                             real_t boxsize,
                             real_t[:] unit_vector_x,
                             real_t[:] unit_vector_y,
                             real_t[:] unit_vector_z,
                             int numthreads):

    cdef int Np = pos.shape[0]
    cdef int ip
    cdef real_t x, y, z
    cdef real_t x_dot_ex, x_dot_ey, x_dot_ez

    cdef int[:] index = np.zeros(Np, dtype=np.intc)

    openmp.omp_set_num_threads(numthreads)

    for ip in prange(Np, nogil=True, schedule='static'):
        x = pos[ip, 0] - xc
        y = pos[ip, 1] - yc
        z = pos[ip, 2] - zc

        # # box_wrap_diff function
        # if x < -0.5*boxsize:
        #     x = x + boxsize
        # elif x > 0.5*boxsize:
        #     x = x - boxsize

        # if y < -0.5*boxsize:
        #     y = y + boxsize
        # elif y > 0.5*boxsize:
        #     y = y - boxsize

        # if z < -0.5*boxsize:
        #     z = z + boxsize
        # elif z > 0.5*boxsize:
        #     z = z - boxsize

        x_dot_ex = x * unit_vector_x[0] + y * unit_vector_x[1] + z * unit_vector_x[2]
        x_dot_ey = x * unit_vector_y[0] + y * unit_vector_y[1] + z * unit_vector_y[2]
        x_dot_ez = x * unit_vector_z[0] + y * unit_vector_z[1] + z * unit_vector_z[2]

        # Index calculation
        index[ip] = 0
        if (x_dot_ex < (sidelength_x + thickness[ip])/2.0) and (x_dot_ex > -(sidelength_x + thickness[ip])/2.0):
            if (x_dot_ey < (sidelength_y + thickness[ip])/2.0) and (x_dot_ey > -(sidelength_y + thickness[ip])/2.0):
                if (x_dot_ez < (sidelength_z + thickness[ip])/2.0):
                    if (x_dot_ez > -(sidelength_z + thickness[ip])/2.0):
                        index[ip] = 1

    # Return a numpy boolean array
    tmp = np.zeros(Np, dtype=np.bool_)
    tmp[:] = index[:]
    return tmp


def get_radial_range(real_t [:, :] pos, real_t xc, real_t yc,
                    real_t zc, real_t r_min, real_t r_max,
                    int numthreads):

    cdef int Np = pos.shape[0]
    cdef int ip
    cdef real_t x, y, z, r2

    cdef real_t r2_min = r_min*r_min
    cdef real_t r2_max = r_max*r_max

    cdef int[:] index = np.zeros(Np, dtype=np.intc)

    openmp.omp_set_num_threads(numthreads)

    for ip in prange(Np, nogil=True, schedule='static'):
        x = pos[ip, 0] - xc
        y = pos[ip, 1] - yc
        z = pos[ip, 2] - zc

        r2 = x*x + y*y + z*z

        # Index calculation
        index[ip] = 0
        if (r2 < r2_max) and (r2 > r2_min):
            index[ip] = 1

    # Return a numpy boolean array
    tmp = np.zeros(Np, dtype=np.bool_)
    tmp[:] = index[:]
    return tmp



"""This defines a reader for Arepo Friends-of-Friends (FoF) and subhalo catalog files"""
import numpy as np
import h5py
from .paicos_readers import PaicosReader
from .. import settings
import numbers
import warnings


class Catalog(PaicosReader):
    """
    This is a Python class for reading Arepo group and subhalo catalogs.
    The class is based on a script originally written by Ewald Puchwein,
    which has since then been modified and included in Paicos.

    The class takes in the path of the directory containing the catalog, the
    catalog number, and an optional basename parameter, and uses this
    information to locate and open the catalog files. The class also loads
    the catalog's header, parameters, and configuration.
    The class also includes methods to extract the redshift, scale
    factor, and other properties of the catalog.

    Important methods and attributes.

    cat = Catalog()

    cat.Parameters (dict): Contains information from the parameter
                            file used in the simulation (e.g. param.txt).

    cat.Config (dict): Contains information from the Config
                        file used in the simulation (e.g. Config.txt).

    cat.Header (dict): Contains information about this particular catalog
                        such as its time (e.g scale factor).

    cat.z (float): redshift

    cat.h (float): reduced Hubble param (e.g. 0.67)

    cat.age: the age of the Universe (only for cosmological runs)
    cat.lookback_time: the age of the Universe (only for cosmological runs)

    """
    def __init__(self, basedir='.', snapnum=None,
                 to_physical=False, subfind_catalog=True, verbose=False):

        if subfind_catalog:
            basename = 'fof_subhalo_tab'
        else:
            basename = 'fof_tab'

        super().__init__(basedir=basedir, snapnum=snapnum, basename=basename,
                         load_all=False, to_physical=to_physical,
                         basesubdir='groups', verbose=verbose)

        # give names to specific header fields
        self.nfiles = self.Header["NumFiles"]
        self.ngroups = self.Header["Ngroups_Total"]

        if "Nsubgroups_Total" in self.Header.keys():
            self.nsubs = self.Header["Nsubgroups_Total"]
        else:
            self.nsubs = self.Header["Nsubhalos_Total"]

        # Load all data
        self.load_all_data()

    def load_data(self):
        """
        Overwrite the base class method
        """
        pass

    def load_all_data(self):
        """
        Calling method simply loads all the data in the catalog.

        For large catalogs it might be useful to implement on-demand
        access in a similar way to what we have for snapshots.
        """

        skip_gr = 0
        skip_sub = 0
        for ifile in range(self.nfiles):
            if self.multi_file is False:
                cur_filename = self.filename
            else:
                if self.no_subdir:
                    cur_filename = self.multi_wo_dir.format(ifile)
                else:
                    cur_filename = self.multi_filename.format(ifile)

            if self.verbose:
                print("reading file", cur_filename)

            file = h5py.File(cur_filename, "r")

            ng = int(file["Header"].attrs["Ngroups_ThisFile"])

            if "Nsubgroups_ThisFile" in file["Header"].attrs.keys():
                ns = int(file["Header"].attrs["Nsubgroups_ThisFile"])
            else:
                ns = int(file["Header"].attrs["Nsubhalos_ThisFile"])

            # initialze arrays
            if ifile == 0:
                self.Group = {}
                self.Sub = {}
                for ikey in file["Group"].keys():
                    if len(file["Group/" + ikey].shape) == 1:
                        self.Group[ikey] = np.empty(
                            self.ngroups, dtype=file["Group/" + ikey].dtype)
                    elif len(file["Group/" + ikey].shape) == 2:
                        self.Group[ikey] = np.empty(
                            (self.ngroups, file["Group/" + ikey].shape[1]),
                            dtype=file["Group/" + ikey].dtype)
                    else:
                        assert False

                for ikey in file["Subhalo"].keys():
                    if len(file["Subhalo/" + ikey].shape) == 1:
                        self.Sub[ikey] = np.empty(
                            self.nsubs, dtype=file["Subhalo/" + ikey].dtype)
                    elif len(file["Subhalo/" + ikey].shape) == 2:
                        self.Sub[ikey] = np.empty(
                            (self.nsubs, file["Subhalo/" + ikey].shape[1]),
                            dtype=file["Subhalo/" + ikey].dtype)
                    else:
                        assert False

            # read group data
            for ikey in file["Group"].keys():
                self.Group[ikey][skip_gr:skip_gr + ng] = file["Group/" + ikey]

            # read subhalo data
            for ikey in file["Subhalo"].keys():
                self.Sub[ikey][skip_sub:skip_sub + ns] = file["Subhalo/" + ikey]

            skip_gr += ng
            skip_sub += ns

            file.close()

        # Load all variables with double precision
        if settings.double_precision:

            for ikey in self.Group:
                if not issubclass(self.Group[ikey].dtype.type, numbers.Integral):
                    self.Group[ikey] = self.Group[ikey].astype(np.float64)

            for ikey in self.Sub:
                if not issubclass(self.Sub[ikey].dtype.type, numbers.Integral):
                    self.Sub[ikey] = self.Sub[ikey].astype(np.float64)
        else:
            warnings.warn('\n\nThe cython routines expect double precision '
                          + 'and will fail unless settings.double_precision '
                          + 'is True.\n\n')

        if settings.use_units:
            for key in list(self.Group.keys()):
                self.Group[key] = self.get_paicos_quantity(
                    self.Group[key], key,
                    field='groups')
                if not hasattr(self.Group[key], 'unit'):
                    del self.Group[key]

            for key in list(self.Sub.keys()):
                self.Sub[key] = self.get_paicos_quantity(
                    self.Sub[key], key,
                    field='subhalos')
                if not hasattr(self.Sub[key], 'unit'):
                    del self.Sub[key]



"""
Defines a class that creates an image of a given variable by projecting it
onto a 2D plane using nested grids.
"""
import numpy as np
from .. import settings
from .projector import Projector
from ..util import remove_astro_units


class NestedProjector(Projector):
    """
    This class implements an SPH-like projection of gas variables using nested
    grids. It is a subclass of the Projector class, and inherits its
    properties. It has additional attributes for handling the nested grid
    projections, and additional methods for dealing with resolution,
    digitizing particles, and summing up the different resolutions of the
    images.
    """

    def __init__(self, snap, center, widths, direction,
                 npix=512, nvol=8, factor=3, npix_min=128,
                 verbose=False, make_snap_with_selection=True,
                 store_subimages=False):
        """
        This is the constructor for the NestedProjector class. It initializes
        the properties inherited from the Projector class, as well as the
        properties specific to nested projections.

        Parameters
        ----------
        snap : Snapshot
            A snapshot object of Snapshot class from paicos package.

        center : numpy array
            Center of the region on which projection is to be done, e.g.
            center = [x_c, y_c, z_c].

        widths : numpy array
            Widths of the region on which projection is to be done,
            e.g.m widths=[width_x, width_y, width_z].

        direction : str
            Direction of the projection, e.g. 'x', 'y' or 'z'.

        npix : int, optional
            Number of pixels in the horizontal direction of the image,
            by default 512.

        nvol : int, optional
            Integer used to determine the smoothing length, by default 8

        factor (int, optional): Multiplicative factor in digitizing, defaults to 3.

        npix_min (int, optional): Minimum number of pixels, defaults to 128.

        verbose (bool, optional): Flag for verbosity, defaults to False.

        make_snap_with_selection (bool, optional):
            a boolean indicating if a new snapshot object should be made with
            the selected region, defaults to False

        store_subimages (bool, optional): Flag for storing sub-images, which
        is useful mainly for testing purposes. Defaults to False.

        """

        super().__init__(snap, center, widths, direction,
                         npix=npix, nvol=nvol,
                         make_snap_with_selection=make_snap_with_selection)

        self.verbose = verbose
        self.store_subimages = store_subimages
        self.factor = factor
        self.npix_min = npix_min

        self._bind_to(self._initialize_nested)

        self._initialize_nested()

    def _initialize_nested(self):

        # Code specific for nested functionality below

        # Find required grid resolutions and the binning in smoothing (hsml)
        bins, n_grids = self._get_bins(self.extent[1] - self.extent[0])

        # Digitize particles (make OpenMP cython version of this?)
        digitize = remove_astro_units(np.digitize)
        i_digit = digitize(self.hsml, bins=bins)
        n_particles = self.hsml.shape[0]
        count = 0
        for ii, n_grid in enumerate(n_grids):
            index = i_digit == (ii + 1)
            count += np.sum(index)
            if self.verbose:
                print(f'n_grid={n_grid} contains {np.sum(index)} particles')

        err_msg = f'n_particles={n_particles}, count={count}, need to include all cells!'
        assert n_particles == count, err_msg

        self.n_grids = n_grids
        self.bins = bins
        self.i_digit = i_digit

    @remove_astro_units
    def _get_bins(self, width):
        """
        This helper method finds the required grid resolutions and the
        binning in smoothing (hsml) for the nested grids.

        Parameters:
        width (float): The width of the projection.

        Returns:
        bins (list): A list of float values representing the binned intervals
                     in smoothing (hsml).

        n_grids (list): A list of integers representing the grid resolutions.
        """

        @remove_astro_units
        def nearest_power_of_two(x):
            return int(2**np.ceil(np.log2(x)))

        def log2int(x):
            return int(np.log2(x))

        npix_low = nearest_power_of_two(
            width / np.max(self.hsml) * self.factor)
        npix_high = nearest_power_of_two(
            width / np.min(self.hsml) * self.factor)

        npix_high = self.npix
        npix_low = max(npix_low, self.npix_min)
        npix_low = min(npix_low, npix_high)

        n_grids = []
        bins = [width]
        for ii in range(log2int(npix_low), log2int(npix_high) + 1):
            n_grid = 2**ii

            if self.verbose:
                print(n_grid, width / n_grid)
            n_grids.append(n_grid)
            if n_grid == npix_high:
                bins.append(0.0)
            else:
                bins.append(width / n_grid * self.factor)

        return bins, n_grids

    def increase_image_resolution(self, image, factor):
        """
        Increase the number of pixes without changing the total 'mass'
        of the image
        """
        if factor == 1:
            return image
        repeats = factor
        new_image = np.repeat(np.repeat(image, repeats, axis=0),
                              repeats, axis=1)
        return new_image / factor**2

    def sum_contributions(self, images):
        """
        Given the list of images at various resolutions, sum them up.
        """
        full_image = np.zeros(images[-1].shape)
        n = images[-1].shape[0]
        for image in images:
            full_image += self.increase_image_resolution(image,
                                                         n // image.shape[0])

        return full_image

    @remove_astro_units
    def _cython_project(self, center, widths, variable):
        """
        This method performs the projection of a given variable onto a 2D
        plane using nested grids and a cython implementation.
        """
        if settings.openMP_has_issues:
            from ..cython.sph_projectors import project_image as project
            from ..cython.sph_projectors import project_oriented_image as project_orie
        else:
            from ..cython.sph_projectors import project_image_omp as project
            from ..cython.sph_projectors import project_oriented_image_omp as project_orie

        x_c, y_c, z_c = center[0], center[1], center[2]
        width_x, width_y, width_z = widths

        boxsize = self.snap.box

        images = []
        for ii, n_grid in enumerate(self.n_grids):
            index_n = self.i_digit == (ii + 1)
            hsml_n = self.hsml[index_n]
            pos_n = self.pos[index_n]
            variable_n = variable[index_n]
            if self.direction == 'x':
                proj_n = project(pos_n[:, 1],
                                 pos_n[:, 2],
                                 variable_n,
                                 hsml_n, n_grid,
                                 y_c, z_c, width_y, width_z,
                                 boxsize, settings.numthreads_reduction)
            elif self.direction == 'y':
                proj_n = project(pos_n[:, 2],
                                 pos_n[:, 0],
                                 variable_n,
                                 hsml_n, n_grid,
                                 z_c, x_c, width_z, width_x,
                                 boxsize, settings.numthreads_reduction)
            elif self.direction == 'z':
                proj_n = project(pos_n[:, 0],
                                 pos_n[:, 1],
                                 variable_n,
                                 hsml_n, n_grid,
                                 x_c, y_c, width_x, width_y,
                                 boxsize, settings.numthreads_reduction)
            elif self.direction == 'orientation':
                unit_vectors = self.orientation.cartesian_unit_vectors

                proj_n = project_orie(pos_n[:, 0],
                                      pos_n[:, 1],
                                      pos_n[:, 2],
                                      variable_n,
                                      hsml_n, n_grid,
                                      x_c, y_c, z_c, width_x, width_y,
                                      boxsize,
                                      unit_vectors['x'],
                                      unit_vectors['y'],
                                      unit_vectors['z'],
                                      settings.numthreads_reduction)
            else:
                raise RuntimeError(f'invalid input for direction={self.direction}')
            images.append(proj_n)

        projection = self.sum_contributions(images)

        if self.store_subimages:
            self.images = images

        return projection



# import numpy as np
import astropy.units as u

"""
Here we specify the units used in all the different fields commonly used
in Arepo simulations.

These units have been gathered by looking at the source code
and cross-referencing with the overview at:
https://www.tng-project.org/data/docs/specifications/
"""

unit_less = u.Unit('')

Coordinates = u.Unit('arepo_length small_a / small_h')
Masses = u.Unit('arepo_mass / small_h')
Velocities = u.Unit('arepo_velocity  small_a^(1/2)')

Potential = u.Unit('arepo_velocity^2 / small_a')

MagneticField = u.Unit('arepo_pressure^(1/2) small_a^-2 small_h')

Volume = Coordinates**3
Density = Masses / Volume
Density = u.Unit('arepo_density small_h2 / small_a3')

default = {
    'Coordinates': Coordinates,
    'Masses': Masses,
    'Velocities': Velocities
}


StarFormationRate = u.Unit('Msun/yr')

voronoi_cells = {
    'AllowRefinement': unit_less,
    'BfieldGradient': MagneticField / Coordinates,
    'CenterOfMass': Coordinates,
    'Coordinates': Coordinates,
    'Density': Density,
    'Volume': Volume,
    'ElectronAbundance': unit_less,
    'EnergyDissipation': 'arepo_energy arepo_time^-1 small_a^-1',
    'GFM_AGNRadiation': False,  # u.Unit('erg s^-1 cm^-2')*u.Unit(str(4*np.pi)),
    'GFM_CoolingRate': 'erg cm^3 s^-1',
    'GFM_Metallicity': unit_less,
    'GFM_Metals': unit_less,
    'GFM_WindDMVelDisp': 'arepo_velocity',
    'GFM_WindHostHaloMass': Masses,
    'HighResGasMass': Masses,
    'InternalEnergy': 'arepo_energy / arepo_mass',
    'Machnumber': unit_less,
    'MagneticField': MagneticField,
    'MagneticFieldDivergence': MagneticField / Coordinates,
    'MagneticFieldDivergenceAlternative': MagneticField / Coordinates,
    'Masses': Masses,
    'NeutralHydrogenAbundance': unit_less,
    'ParticleIDs': unit_less,
    'Potential': Potential,
    'Pressure': 'arepo_pressure  small_h^2 small_a^-3',
    'StarFormationRate': StarFormationRate,
    'SubfindDMDensity': Density,
    'SubfindDensity': Density,
    'SubfindHsml': Coordinates,
    'SubfindVelDisp': 'arepo_velocity',
    'Velocities': Velocities,
    'VelocityGradient': Velocities / Coordinates
}

dark_matter = {
    'Coordinates': Coordinates,
    'Masses': Masses,
    'ParticleIDs': unit_less,
    'Potential': Potential,
    'SubfindDMDensity': Density,
    'SubfindDensity': Density,
    'SubfindHsml': Coordinates,
    'SubfindVelDisp': 'arepo_velocity',
    'Velocities': Velocities
}

stars = {
    'Coordinates': Coordinates,
    'GFM_InitialMass': Masses,
    'GFM_Metallicity': unit_less,
    'GFM_Metals': unit_less,
    'GFM_StellarFormationTime': unit_less,
    'GFM_StellarPhotometrics': 'mag',
    'Masses': Masses,
    'ParticleIDs': unit_less,
    'Potential': Potential,
    'SubfindDMDensity': Density,
    'SubfindDensity': Density,
    'SubfindHsml': Coordinates,
    'SubfindVelDisp': 'arepo_velocity',
    'Velocities': Velocities
}


energy_inj = u.Unit('small_a^2 small_h^-1 arepo_energy')
Mdot = u.Unit('arepo_mass / arepo_time')

black_holes = {
    'BH_CumEgyInjection_QM': energy_inj,
    'BH_CumEgyInjection_RM': energy_inj,
    'BH_CumMassGrowth_QM': Masses,
    'BH_CumMassGrowth_RM': Masses,
    'BH_Density': Density,
    'BH_HostHaloMass': Masses,
    'BH_Hsml': Coordinates,
    'BH_MPB_CumEgyHigh': energy_inj,
    'BH_MPB_CumEgyLow': energy_inj,
    'BH_Mass': Masses,
    'BH_Mass_bubbles': Masses,
    'BH_Mass_ini': Masses,
    'BH_Mdot': Mdot,
    'BH_MdotBondi': Mdot,
    'BH_MdotEddington': Mdot,
    'BH_Pressure': 'arepo_mass arepo_length^-1 arepo_time^-2',
    'BH_Progs': unit_less,
    'BH_U': 'arepo_velocity^2',
    'Coordinates': Coordinates,
    'Masses': Masses,
    'ParticleIDs': unit_less,
    'Potential': Potential,
    'SubfindDMDensity': Density,
    'SubfindDensity': Density,
    'SubfindHsml': Coordinates,
    'SubfindVelDisp': 'arepo_velocity',
    'Velocities': Velocities
}

groups = {
    'GroupBHMass': Masses,
    'GroupBHMdot': Mdot,
    'GroupCM': Coordinates,
    'GroupFirstSub': unit_less,
    'GroupGasMetalFractions': unit_less,
    'GroupGasMetallicity': unit_less,
    'GroupLen': unit_less,
    'GroupLenType': unit_less,
    'GroupMass': Masses,
    'GroupMassType': Masses,
    'GroupNsubs': unit_less,
    'GroupPos': Coordinates,
    'GroupSFR': StarFormationRate,
    'GroupStarMetalFractions': unit_less,
    'GroupStarMetallicity': unit_less,
    'GroupVel': 'arepo_velocity / small_a',  # note the a-factor!
    'GroupWindMass': Masses,
    'Group_M_Crit200': Masses,
    'Group_M_Crit500': Masses,
    'Group_M_Mean200': Masses,
    'Group_M_TopHat200': Masses,
    'Group_R_Crit200': Coordinates,
    'Group_R_Crit500': Coordinates,
    'Group_R_Mean200': Coordinates,
    'Group_R_TopHat200': Coordinates
}

subhalos = {
    'SubhaloBHMass': Masses,
    'SubhaloBHMdot': Mdot,
    'SubhaloBfldDisk': MagneticField,
    'SubhaloBfldHalo': MagneticField,
    'SubhaloCM': Coordinates,
    'SubhaloGasMetalFractions': unit_less,
    'SubhaloGasMetalFractionsHalfRad': unit_less,
    'SubhaloGasMetalFractionsMaxRad': unit_less,
    'SubhaloGasMetalFractionsSfr': unit_less,
    'SubhaloGasMetalFractionsSfrWeighted': unit_less,
    'SubhaloGasMetallicity': unit_less,
    'SubhaloGasMetallicityHalfRad': unit_less,
    'SubhaloGasMetallicityMaxRad': unit_less,
    'SubhaloGasMetallicitySfr': unit_less,
    'SubhaloGasMetallicitySfrWeighted': unit_less,
    'SubhaloGrNr': unit_less,
    'SubhaloHalfmassRad': Coordinates,
    'SubhaloHalfmassRadType': Coordinates,
    'SubhaloIDMostbound': unit_less,
    'SubhaloLen': unit_less,
    'SubhaloLenType': unit_less,
    'SubhaloMass': Masses,
    'SubhaloMassInHalfRad': Masses,
    'SubhaloMassInHalfRadType': Masses,
    'SubhaloMassInMaxRad': Masses,
    'SubhaloMassInMaxRadType': Masses,
    'SubhaloMassInRad': Masses,
    'SubhaloMassInRadType': Masses,
    'SubhaloMassType': Masses,
    'SubhaloParent': unit_less,
    'SubhaloPos': Coordinates,
    'SubhaloSFR': StarFormationRate,
    'SubhaloSFRinHalfRad': StarFormationRate,
    'SubhaloSFRinMaxRad': StarFormationRate,
    'SubhaloSFRinRad': StarFormationRate,
    'SubhaloSpin': 'arepo_length / small_h arepo_velocity',  # Check!
    'SubhaloStarMetalFractions': unit_less,
    'SubhaloStarMetalFractionsHalfRad': unit_less,
    'SubhaloStarMetalFractionsMaxRad': unit_less,
    'SubhaloStarMetallicity': unit_less,
    'SubhaloStarMetallicityHalfRad': unit_less,
    'SubhaloStarMetallicityMaxRad': unit_less,
    'SubhaloStellarPhotometrics': 'mag',
    'SubhaloStellarPhotometricsMassInRad': Masses,
    'SubhaloStellarPhotometricsRad': Coordinates,
    'SubhaloVel': 'arepo_velocity',
    'SubhaloVelDisp': 'arepo_velocity',
    'SubhaloVmax': 'arepo_velocity',
    'SubhaloVmaxRad': Velocities,
    'SubhaloWindMass': Masses
}

unit_dict = {'default': default,
             'voronoi_cells': voronoi_cells,
             'dark_matter': dark_matter,
             'stars': stars,
             'black_holes': black_holes,
             'groups': groups,
             'subhalos': subhalos
             }



# Boolean determining whether we use Paicos quantities as a default
use_units = True

# Number of threads to use in calculations
numthreads = 16

# Settings for automatic calculation of derived variables
use_only_user_functions = False
print_info_when_deriving_variables = True

# Load all fields as double precision arrays (FALSE is currently not supported
# by the Cython routines, so True is strongly recommended)
double_precision = True

# OpenMP info
give_openMP_warnings = True

# Whether to use aliases, e.g. dens instead of 0_Density
use_aliases = False

# Strictly enforce units
strict_units = True

# Enable spherical coordinates
spherical_coords = False

# Enable cylindrical coordinates
cyl_coords = False

# Whether to load GPU/cuda functionality on startup
load_cuda_functionality_on_startup = True
